from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from math import cos, pi, sin, tan
from typing import (
    TYPE_CHECKING,
    Any,
    ContextManager,
    Literal,
    NoReturn,
    Protocol,
)

from travertino.colors import Color

import toga
from toga.colors import BLACK, color as parse_color
from toga.constants import Baseline, FillRule
from toga.fonts import (
    SYSTEM,
    SYSTEM_DEFAULT_FONT_SIZE,
    Font,
)
from toga.handlers import wrapped_handler

from .base import StyleT, Widget

if TYPE_CHECKING:
    from toga.images import ImageT

#######################################################################################
# Simple drawing objects
#######################################################################################


class DrawingObject(ABC):
    """A drawing operation in a :any:`Context`.

    Every context drawing method creates a ``DrawingObject``, adds it to the context,
    and returns it. Each argument passed to the method becomes a property of the
    ``DrawingObject``, which can be modified as shown in the `Usage`_ section.

    ``DrawingObjects`` can also be created manually, then added to a context using the
    :meth:`~Context.append` or :meth:`~Context.insert` methods. Their constructors take
    the same arguments as the corresponding :any:`Context` method, and their classes
    have the same names, but capitalized:

    * :meth:`toga.widgets.canvas.Arc <Context.arc>`
    * :meth:`toga.widgets.canvas.BeginPath <Context.begin_path>`
    * :meth:`toga.widgets.canvas.BezierCurveTo <Context.bezier_curve_to>`
    * :meth:`toga.widgets.canvas.ClosePath <Context.close_path>`
    * :meth:`toga.widgets.canvas.Ellipse <Context.ellipse>`
    * :meth:`toga.widgets.canvas.Fill <Context.fill>`
    * :meth:`toga.widgets.canvas.LineTo <Context.line_to>`
    * :meth:`toga.widgets.canvas.MoveTo <Context.move_to>`
    * :meth:`toga.widgets.canvas.QuadraticCurveTo <Context.quadratic_curve_to>`
    * :meth:`toga.widgets.canvas.Rect <Context.rect>`
    * :meth:`toga.widgets.canvas.ResetTransform <Context.reset_transform>`
    * :meth:`toga.widgets.canvas.Rotate <Context.rotate>`
    * :meth:`toga.widgets.canvas.Scale <Context.scale>`
    * :meth:`toga.widgets.canvas.Stroke <Context.stroke>`
    * :meth:`toga.widgets.canvas.Translate <Context.translate>`
    * :meth:`toga.widgets.canvas.WriteText <Context.write_text>`
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    @abstractmethod
    def _draw(self, impl: Any, **kwargs: Any) -> None: ...


class BeginPath(DrawingObject):
    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.begin_path(**kwargs)


class ClosePath(DrawingObject):
    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.close_path(**kwargs)


class Fill(DrawingObject):
    def __init__(
        self,
        color: str = BLACK,
        fill_rule: FillRule = FillRule.NONZERO,
    ):
        super().__init__()
        self.color = color
        self.fill_rule = fill_rule

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(color={self.color!r}, "
            f"fill_rule={self.fill_rule})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.fill(self.color, self.fill_rule, **kwargs)

    @property
    def fill_rule(self) -> FillRule:
        return self._fill_rule

    @fill_rule.setter
    def fill_rule(self, fill_rule: FillRule) -> None:
        self._fill_rule = fill_rule

    @property
    def color(self) -> Color:
        return self._color

    @color.setter
    def color(self, value: Color | str | None) -> None:
        if value is None:
            self._color = parse_color(BLACK)
        else:
            self._color = parse_color(value)

    ###########################################################################
    # 2023-07 Backwards incompatibility
    ###########################################################################

    # `context.fill()` used to be a context manager, but is now a primitive.
    # If you try to use the Fill drawing object as a context, raise an exception.
    def __enter__(self) -> NoReturn:
        raise RuntimeError("Context.fill() has been renamed Context.Fill().")

    def __exit__(self) -> None:  # pragma: no cover
        # This method is required to make the object a context manager, but as the
        # __enter__ method raises an exception, the __exit__ can't be called.
        pass


class Stroke(DrawingObject):
    def __init__(
        self,
        color: Color | str | None = BLACK,
        line_width: float = 2.0,
        line_dash: list[float] | None = None,
    ):
        super().__init__()
        self.color = color
        self.line_width = line_width
        self.line_dash = line_dash

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(color={self.color!r}, "
            f"line_width={self.line_width}, line_dash={self.line_dash!r})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.stroke(self.color, self.line_width, self.line_dash, **kwargs)

    @property
    def color(self) -> Color:
        return self._color

    @color.setter
    def color(self, value: Color | str | None) -> None:
        if value is None:
            self._color = parse_color(BLACK)
        else:
            self._color = parse_color(value)

    @property
    def line_width(self) -> float:
        return self._line_width

    @line_width.setter
    def line_width(self, value: float) -> None:
        self._line_width = float(value)

    @property
    def line_dash(self) -> list[float] | None:
        return self._line_dash

    @line_dash.setter
    def line_dash(self, value: list[float] | None) -> None:
        self._line_dash = value

    ###########################################################################
    # 2023-07 Backwards incompatibility
    ###########################################################################

    # `context.stroke()` used to be a context manager, but is now a primitive.
    # If you try to use the Stroke drawing object as a context, raise an exception.
    def __enter__(self) -> NoReturn:
        raise RuntimeError("Context.stroke() has been renamed Context.Stroke().")

    def __exit__(self) -> None:  # pragma: no cover
        # This method is required to make the object a context manager, but as the
        # __enter__ method raises an exception, the __exit__ can't be called.
        pass


class MoveTo(DrawingObject):
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.x}, y={self.y})"

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.move_to(self.x, self.y, **kwargs)


class LineTo(DrawingObject):
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.x}, y={self.y})"

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.line_to(self.x, self.y, **kwargs)


class BezierCurveTo(DrawingObject):
    def __init__(
        self, cp1x: float, cp1y: float, cp2x: float, cp2y: float, x: float, y: float
    ):
        self.cp1x = cp1x
        self.cp1y = cp1y
        self.cp2x = cp2x
        self.cp2y = cp2y
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(cp1x={self.cp1x}, cp1y={self.cp1y}, "
            f"cp2x={self.cp2x}, cp2y={self.cp2y}, "
            f"x={self.x}, y={self.y})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.bezier_curve_to(
            self.cp1x, self.cp1y, self.cp2x, self.cp2y, self.x, self.y, **kwargs
        )


class QuadraticCurveTo(DrawingObject):
    def __init__(self, cpx: float, cpy: float, x: float, y: float):
        self.cpx = cpx
        self.cpy = cpy
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(cpx={self.cpx}, cpy={self.cpy}, x={self.x}, y={self.y})"

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.quadratic_curve_to(self.cpx, self.cpy, self.x, self.y, **kwargs)


class Arc(DrawingObject):
    def __init__(
        self,
        x: float,
        y: float,
        radius: float,
        startangle: float = 0.0,
        endangle: float = 2 * pi,
        anticlockwise: bool = False,
    ):
        self.x = x
        self.y = y
        self.radius = radius
        self.startangle = startangle
        self.endangle = endangle
        self.anticlockwise = anticlockwise

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.x}, y={self.y}, "
            f"radius={self.radius}, startangle={self.startangle:.3f}, "
            f"endangle={self.endangle:.3f}, anticlockwise={self.anticlockwise})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.arc(
            self.x,
            self.y,
            self.radius,
            self.startangle,
            self.endangle,
            self.anticlockwise,
            **kwargs,
        )


class Ellipse(DrawingObject):
    def __init__(
        self,
        x: float,
        y: float,
        radiusx: float,
        radiusy: float,
        rotation: float = 0.0,
        startangle: float = 0.0,
        endangle: float = 2 * pi,
        anticlockwise: bool = False,
    ):
        self.x = x
        self.y = y
        self.radiusx = radiusx
        self.radiusy = radiusy
        self.rotation = rotation
        self.startangle = startangle
        self.endangle = endangle
        self.anticlockwise = anticlockwise

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.x}, y={self.y}, "
            f"radiusx={self.radiusx}, radiusy={self.radiusy}, "
            f"rotation={self.rotation:.3f}, startangle={self.startangle:.3f}, "
            f"endangle={self.endangle:.3f}, anticlockwise={self.anticlockwise})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.ellipse(
            self.x,
            self.y,
            self.radiusx,
            self.radiusy,
            self.rotation,
            self.startangle,
            self.endangle,
            self.anticlockwise,
            **kwargs,
        )


class Rect(DrawingObject):
    def __init__(self, x: float, y: float, width: float, height: float):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.x}, y={self.y}, "
            f"width={self.width}, height={self.height})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.rect(self.x, self.y, self.width, self.height, **kwargs)


class WriteText(DrawingObject):
    def __init__(
        self,
        text: str,
        x: float = 0.0,
        y: float = 0.0,
        font: Font | None = None,
        baseline: Baseline = Baseline.ALPHABETIC,
    ):
        self.text = text
        self.x = x
        self.y = y
        self.font = font
        self.baseline = baseline

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(text={self.text!r}, x={self.x}, y={self.y}, "
            f"font={self.font!r}, baseline={self.baseline})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.write_text(
            str(self.text), self.x, self.y, self.font._impl, self.baseline, **kwargs
        )

    @property
    def font(self) -> Font:
        return self._font

    @font.setter
    def font(self, value: Font | None) -> None:
        if value is None:
            self._font = Font(family=SYSTEM, size=SYSTEM_DEFAULT_FONT_SIZE)
        else:
            self._font = value


class Rotate(DrawingObject):
    def __init__(self, radians: float):
        self.radians = radians

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(radians={self.radians:.3f})"

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.rotate(self.radians, **kwargs)


class Scale(DrawingObject):
    def __init__(self, sx: float, sy: float):
        self.sx = sx
        self.sy = sy

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(sx={self.sx:.3f}, sy={self.sy:.3f})"

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.scale(self.sx, self.sy, **kwargs)


class Translate(DrawingObject):
    def __init__(self, tx: float, ty: float):
        self.tx = tx
        self.ty = ty

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(tx={self.tx}, ty={self.ty})"

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.translate(self.tx, self.ty, **kwargs)


class ResetTransform(DrawingObject):
    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.reset_transform(**kwargs)


#######################################################################################
# Drawing Contexts
#######################################################################################


class Context(DrawingObject):
    """A drawing context for a canvas.

    You should not create a :class:`~toga.widgets.canvas.Context` directly; instead,
    you should use the :meth:`~Context.Context` method on an existing context,
    or use :any:`Canvas.context` to access the root context of the canvas.
    """

    def __init__(self, canvas: toga.Canvas, **kwargs: Any):
        # kwargs used to support multiple inheritance
        super().__init__(**kwargs)
        self._canvas = canvas
        self.drawing_objects: list[DrawingObject] = []

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.push_context(**kwargs)
        for obj in self.drawing_objects:
            obj._draw(impl, **kwargs)
        impl.pop_context(**kwargs)

    ###########################################################################
    # Methods to keep track of the canvas, automatically redraw it
    ###########################################################################

    @property
    def canvas(self) -> Canvas:
        """The canvas that is associated with this drawing context."""
        return self._canvas

    def redraw(self) -> None:
        """Calls :any:`Canvas.redraw` on the parent Canvas."""
        self.canvas.redraw()

    ###########################################################################
    # Operations on drawing objects
    ###########################################################################

    def __len__(self) -> int:
        """Returns the number of drawing objects that are in this context."""
        return len(self.drawing_objects)

    def __getitem__(self, index: int) -> DrawingObject:
        """Returns the drawing object at the given index."""
        return self.drawing_objects[index]

    def append(self, obj: DrawingObject) -> None:
        """Append a drawing object to the context.

        :param obj: The drawing object to add to the context.
        """
        self.drawing_objects.append(obj)
        self.redraw()

    def insert(self, index: int, obj: DrawingObject) -> None:
        """Insert a drawing object into the context at a specific index.

        :param index: The index at which the drawing object should be inserted.
        :param obj: The drawing object to add to the context.
        """
        self.drawing_objects.insert(index, obj)
        self.redraw()

    def remove(self, obj: DrawingObject) -> None:
        """Remove a drawing object from the context.

        :param obj: The drawing object to remove.
        """
        self.drawing_objects.remove(obj)
        self.redraw()

    def clear(self) -> None:
        """Remove all drawing objects from the context."""
        self.drawing_objects.clear()
        self.redraw()

    ###########################################################################
    # Path manipulation
    ###########################################################################

    def begin_path(self) -> BeginPath:
        """Start a new path in the canvas context.

        :returns: The ``BeginPath`` :any:`DrawingObject` for the operation.
        """
        begin_path = BeginPath()
        self.append(begin_path)
        return begin_path

    def close_path(self) -> ClosePath:
        """Close the current path in the canvas context.

        This closes the current path as a simple drawing operation. It should be paired
        with a :meth:`~toga.widgets.canvas.Context.begin_path()` operation; or, to
        complete a complete closed path, use the
        :meth:`~toga.widgets.canvas.Context.ClosedPath()` context manager.

        :returns: The ``ClosePath`` :any:`DrawingObject` for the operation.
        """
        close_path = ClosePath()
        self.append(close_path)
        return close_path

    def move_to(self, x: float, y: float) -> MoveTo:
        """Moves the current point of the canvas context without drawing.

        :param x: The x coordinate of the new current point.
        :param y: The y coordinate of the new current point.
        :returns: The ``MoveTo`` :any:`DrawingObject` for the operation.
        """
        move_to = MoveTo(x, y)
        self.append(move_to)
        return move_to

    def line_to(self, x: float, y: float) -> LineTo:
        """Draw a line segment ending at a point in the canvas context.

        :param x: The x coordinate for the end point of the line segment.
        :param y: The y coordinate for the end point of the line segment.
        :returns: The ``LineTo`` :any:`DrawingObject` for the operation.
        """
        line_to = LineTo(x, y)
        self.append(line_to)
        return line_to

    def bezier_curve_to(
        self,
        cp1x: float,
        cp1y: float,
        cp2x: float,
        cp2y: float,
        x: float,
        y: float,
    ) -> BezierCurveTo:
        """Draw a Bézier curve in the canvas context.

        A Bézier curve requires three points. The first two are control points; the
        third is the end point for the curve. The starting point is the last point in
        the current path, which can be changed using move_to() before creating the
        Bézier curve.

        :param cp1y: The y coordinate for the first control point of the Bézier curve.
        :param cp1x: The x coordinate for the first control point of the Bézier curve.
        :param cp2x: The x coordinate for the second control point of the Bézier curve.
        :param cp2y: The y coordinate for the second control point of the Bézier curve.
        :param x: The x coordinate for the end point.
        :param y: The y coordinate for the end point.
        :returns: The ``BezierCurveTo`` :any:`DrawingObject` for the operation.
        """
        bezier_curve_to = BezierCurveTo(cp1x, cp1y, cp2x, cp2y, x, y)
        self.append(bezier_curve_to)
        return bezier_curve_to

    def quadratic_curve_to(
        self,
        cpx: float,
        cpy: float,
        x: float,
        y: float,
    ) -> QuadraticCurveTo:
        """Draw a quadratic curve in the canvas context.

        A quadratic curve requires two points. The first point is a control point; the
        second is the end point. The starting point of the curve is the last point in
        the current path, which can be changed using ``moveTo()`` before creating the
        quadratic curve.

        :param cpx: The x axis of the coordinate for the control point of the quadratic
            curve.
        :param cpy: The y axis of the coordinate for the control point of the quadratic
            curve.
        :param x: The x axis of the coordinate for the end point.
        :param y: The y axis of the coordinate for the end point.
        :returns: The ``QuadraticCurveTo`` :any:`DrawingObject` for the operation.
        """
        quadratic_curve_to = QuadraticCurveTo(cpx, cpy, x, y)
        self.append(quadratic_curve_to)
        return quadratic_curve_to

    def arc(
        self,
        x: float,
        y: float,
        radius: float,
        startangle: float = 0.0,
        endangle: float = 2 * pi,
        anticlockwise: bool = False,
    ) -> Arc:
        """Draw a circular arc in the canvas context.

        A full circle will be drawn by default; an arc can be drawn by specifying a
        start and end angle.

        :param x: The X coordinate of the circle's center.
        :param y: The Y coordinate of the circle's center.
        :param startangle: The start angle in radians, measured clockwise from the
            positive X axis.
        :param endangle: The end angle in radians, measured clockwise from the positive
            X axis.
        :param anticlockwise: If true, the arc is swept anticlockwise. The default is
            clockwise.
        :returns: The ``Arc`` :any:`DrawingObject` for the operation.
        """
        arc = Arc(x, y, radius, startangle, endangle, anticlockwise)
        self.append(arc)
        return arc

    def ellipse(
        self,
        x: float,
        y: float,
        radiusx: float,
        radiusy: float,
        rotation: float = 0.0,
        startangle: float = 0.0,
        endangle: float = 2 * pi,
        anticlockwise: bool = False,
    ) -> Ellipse:
        """Draw an elliptical arc in the canvas context.

        A full ellipse will be drawn by default; an arc can be drawn by specifying a
        start and end angle.

        :param x: The X coordinate of the ellipse's center.
        :param y: The Y coordinate of the ellipse's center.
        :param radiusx: The ellipse's horizontal axis radius.
        :param radiusy: The ellipse's vertical axis radius.
        :param rotation: The ellipse's rotation in radians, measured clockwise around
            its center.
        :param startangle: The start angle in radians, measured clockwise from the
            positive X axis.
        :param endangle: The end angle in radians, measured clockwise from the positive
            X axis.
        :param anticlockwise: If true, the arc is swept anticlockwise. The default is
            clockwise.
        :returns: The ``Ellipse`` :any:`DrawingObject` for the operation.
        """
        ellipse = Ellipse(
            x,
            y,
            radiusx,
            radiusy,
            rotation,
            startangle,
            endangle,
            anticlockwise,
        )
        self.append(ellipse)
        return ellipse

    def rect(self, x: float, y: float, width: float, height: float) -> Rect:
        """Draw a rectangle in the canvas context.

        :param x: The horizontal coordinate of the left of the rectangle.
        :param y: The vertical coordinate of the top of the rectangle.
        :param width: The width of the rectangle.
        :param height: The height of the rectangle.
        :returns: The ``Rect`` :any:`DrawingObject` for the operation.
        """
        rect = Rect(x, y, width, height)
        self.append(rect)
        return rect

    def fill(
        self,
        color: str = BLACK,
        fill_rule: FillRule = FillRule.NONZERO,
        preserve: None = None,  # DEPRECATED
    ) -> Fill:
        """Fill the current path.

        The fill can use either the `Non-Zero
        <https://en.wikipedia.org/wiki/Nonzero-rule>`__ or `Even-Odd
        <https://en.wikipedia.org/wiki/Even-odd_rule>`__ winding rule for filling paths.

        :param fill_rule: `nonzero` is the non-zero winding rule; `evenodd` is the
            even-odd winding rule.
        :param color: The fill color.
        :param preserve: **DEPRECATED**: this argument has no effect.
        :returns: The ``Fill`` :any:`DrawingObject` for the operation.
        """
        if preserve is not None:
            warnings.warn(
                "The `preserve` argument on fill() has been deprecated.",
                DeprecationWarning,
            )

        fill = Fill(color, fill_rule)
        self.append(fill)
        return fill

    def stroke(
        self,
        color: str = BLACK,
        line_width: float = 2.0,
        line_dash: list[float] | None = None,
    ) -> Stroke:
        """Draw the current path as a stroke.

        :param color: The color for the stroke.
        :param line_width: The width of the stroke.
        :param line_dash: The dash pattern to follow when drawing the line, expressed as
            alternating lengths of dashes and spaces. The default is a solid line.
        :returns: The ``Stroke`` :any:`DrawingObject` for the operation.
        """
        stroke = Stroke(color, line_width, line_dash)
        self.append(stroke)
        return stroke

    ###########################################################################
    # Text drawing
    ###########################################################################

    def write_text(
        self,
        text: str,
        x: float = 0.0,
        y: float = 0.0,
        font: Font | None = None,
        baseline: Baseline = Baseline.ALPHABETIC,
    ) -> WriteText:
        """Write text at a given position in the canvas context.

        Drawing text is effectively a series of path operations, so the text will have
        the color and fill properties of the canvas context.

        :param text: The text to draw. Newlines will cause line breaks, but long lines
            will not be wrapped.
        :param x: The X coordinate of the text's left edge.
        :param y: The Y coordinate: its meaning depends on ``baseline``.
        :param font: The font in which to draw the text. The default is the system font.
        :param baseline: Alignment of text relative to the Y coordinate.
        :returns: The ``WriteText`` :any:`DrawingObject` for the operation.
        """
        write_text = WriteText(text, x, y, font, baseline)
        self.append(write_text)
        return write_text

    ###########################################################################
    # Transformations
    ###########################################################################
    def rotate(self, radians: float) -> Rotate:
        """Add a rotation to the canvas context.

        :param radians: The angle to rotate clockwise in radians.
        :returns: The ``Rotate`` :any:`DrawingObject` for the transformation.
        """
        rotate = Rotate(radians)
        self.append(rotate)
        return rotate

    def scale(self, sx: float, sy: float) -> Scale:
        """Add a scaling transformation to the canvas context.

        :param sx: Scale factor for the X dimension. A negative value flips the
            image horizontally.
        :param sy: Scale factor for the Y dimension. A negative value flips the
            image vertically.
        :returns: The ``Scale`` :any:`DrawingObject` for the transformation.
        """
        scale = Scale(sx, sy)
        self.append(scale)
        return scale

    def translate(self, tx: float, ty: float) -> Translate:
        """Add a translation to the canvas context.

        :param tx: Translation for the X dimension.
        :param ty: Translation for the Y dimension.
        :returns: The ``Translate`` :any:`DrawingObject` for the transformation.
        """
        translate = Translate(tx, ty)
        self.append(translate)
        return translate

    def reset_transform(self) -> ResetTransform:
        """Reset all transformations in the canvas context.

        :returns: A ``ResetTransform`` :any:`DrawingObject`.
        """
        reset_transform = ResetTransform()
        self.append(reset_transform)
        return reset_transform

    ###########################################################################
    # Subcontexts of this context
    ###########################################################################

    @contextmanager
    def Context(self) -> Iterator[Context]:
        """Construct and yield a new sub-:class:`~toga.widgets.canvas.Context` within
        this context.

        :yields: The new :class:`~toga.widgets.canvas.Context` object.
        """
        context = Context(canvas=self._canvas)
        self.append(context)
        yield context
        self.redraw()

    @contextmanager
    def ClosedPath(
        self,
        x: float | None = None,
        y: float | None = None,
    ) -> Iterator[ClosedPathContext]:
        """Construct and yield a new :class:`~toga.widgets.canvas.ClosedPath`
        sub-context that will draw a closed path, starting from an origin.

        This is a context manager; it creates a new path and moves to the start
        coordinate; when the context exits, the path is closed. For fine-grained control
        of a path, you can use :meth:`~toga.widgets.canvas.Context.begin_path` and
        :meth:`~toga.widgets.canvas.Context.close_path`.

        :param x: The x coordinate of the path's starting point.
        :param y: The y coordinate of the path's starting point.
        :yields: The :class:`~toga.widgets.canvas.ClosedPathContext` context object.
        """
        closed_path = ClosedPathContext(canvas=self.canvas, x=x, y=y)
        self.append(closed_path)
        yield closed_path

    @contextmanager
    def Fill(
        self,
        x: float | None = None,
        y: float | None = None,
        color: str = BLACK,
        fill_rule: FillRule = FillRule.NONZERO,
    ) -> Iterator[FillContext]:
        """Construct and yield a new :class:`~toga.widgets.canvas.Fill` sub-context
        within this context.

        This is a context manager; it creates a new path, and moves to the start
        coordinate; when the context exits, the path is closed with a fill. For
        fine-grained control of a path, you can use
        :class:`~toga.widgets.canvas.Context.begin_path`,
        :class:`~toga.widgets.canvas.Context.move_to`,
        :class:`~toga.widgets.canvas.Context.close_path` and
        :class:`~toga.widgets.canvas.Context.fill`.

        If both an x and y coordinate is provided, the drawing context will begin with
        a ``move_to`` operation in that context.

        :param x: The x coordinate of the path's starting point.
        :param y: The y coordinate of the path's starting point.
        :param fill_rule: `nonzero` is the non-zero winding rule; `evenodd` is the
            even-odd winding rule.
        :param color: The fill color.
        :yields: The new :class:`~toga.widgets.canvas.FillContext` context object.
        """
        fill = FillContext(
            canvas=self.canvas,
            x=x,
            y=y,
            color=color,
            fill_rule=fill_rule,
        )
        self.append(fill)
        yield fill

    @contextmanager
    def Stroke(
        self,
        x: float | None = None,
        y: float | None = None,
        color: str = BLACK,
        line_width: float = 2.0,
        line_dash: list[float] | None = None,
    ) -> Iterator[StrokeContext]:
        """Construct and yield a new :class:`~toga.widgets.canvas.Stroke` sub-context
        within this context.

        This is a context manager; it creates a new path, and moves to the start
        coordinate; when the context exits, the path is closed with a stroke. For
        fine-grained control of a path, you can use
        :class:`~toga.widgets.canvas.Context.begin_path`,
        :class:`~toga.widgets.canvas.Context.move_to`,
        :class:`~toga.widgets.canvas.Context.close_path` and
        :class:`~toga.widgets.canvas.Context.stroke`.

        If both an x and y coordinate is provided, the drawing context will begin with
        a ``move_to`` operation in that context.

        :param x: The x coordinate of the path's starting point.
        :param y: The y coordinate of the path's starting point.
        :param color: The color for the stroke.
        :param line_width: The width of the stroke.
        :param line_dash: The dash pattern to follow when drawing the line. Default is a
            solid line.
        :yields: The new :class:`~toga.widgets.canvas.StrokeContext` context object.
        """
        stroke = StrokeContext(
            canvas=self.canvas,
            x=x,
            y=y,
            color=color,
            line_width=line_width,
            line_dash=line_dash,
        )
        self.append(stroke)
        yield stroke

    ###########################################################################
    # 2023-07 Backwards incompatibility
    ###########################################################################

    def new_path(self) -> BeginPath:
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.begin_path`."""
        warnings.warn(
            "Context.new_path() has been renamed Context.begin_path()",
            DeprecationWarning,
        )
        return self.begin_path()

    def context(self):
        """**DEPRECATED** - use :meth:`~toga.widgets.canvas.Context.Context`"""
        warnings.warn(
            "Context.context() has been renamed Context.Context()", DeprecationWarning
        )
        return self.Context()

    def closed_path(self, x: float, y: float):
        """**DEPRECATED** - use :meth:`~toga.widgets.canvas.Context.ClosedPath`"""
        warnings.warn(
            "Context.closed_path() has been renamed Context.ClosedPath()",
            DeprecationWarning,
        )
        return self.ClosedPath(x, y)


class ClosedPathContext(Context):
    """A drawing context that will build a closed path, starting from an
    origin.

    This is a context manager; it creates a new path and moves to the start coordinate;
    when the context exits, the path is closed. For fine-grained control of a path, you
    can use :class:`~toga.widgets.canvas.Context.begin_path`,
    :class:`~toga.widgets.canvas.Context.move_to` and
    :class:`~toga.widgets.canvas.Context.close_path`.

    If both an x and y coordinate is provided, the drawing context will begin with
    a ``move_to`` operation in that context.

    You should not create a :class:`~toga.widgets.canvas.ClosedPathContext` context
    directly; instead, you should use the
    :meth:`~toga.widgets.canvas.Context.ClosedPath` method on an existing context.
    """

    def __init__(
        self,
        canvas: toga.Canvas,
        x: float | None = None,
        y: float | None = None,
    ):
        super().__init__(canvas=canvas)
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.x}, y={self.y})"

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        """Used by parent to draw all objects that are part of the context."""
        impl.push_context(**kwargs)
        impl.begin_path(**kwargs)
        if self.x is not None and self.y is not None:
            impl.move_to(x=self.x, y=self.y, **kwargs)

        sub_kwargs = kwargs.copy()
        for obj in self.drawing_objects:
            obj._draw(impl, **sub_kwargs)

        impl.close_path(**kwargs)
        impl.pop_context(**kwargs)


class FillContext(ClosedPathContext):
    """A drawing context that will apply a fill to any paths all objects in the
    context.

    The fill can use either the `Non-Zero
    <https://en.wikipedia.org/wiki/Nonzero-rule>`__ or `Even-Odd
    <https://en.wikipedia.org/wiki/Even-odd_rule>`__ winding rule for filling paths.

    This is a context manager; it creates a new path, and moves to the start coordinate;
    when the context exits, the path is closed with a fill. For fine-grained control of
    a path, you can use :class:`~toga.widgets.canvas.Context.begin_path`,
    :class:`~toga.widgets.canvas.Context.move_to`,
    :class:`~toga.widgets.canvas.Context.close_path` and
    :class:`~toga.widgets.canvas.Context.fill`.

    If both an x and y coordinate is provided, the drawing context will begin with
    a ``move_to`` operation in that context.

    You should not create a :class:`~toga.widgets.canvas.FillContext` context directly;
    instead, you should use the :meth:`~toga.widgets.canvas.Context.Fill` method on an
    existing context.
    """

    def __init__(
        self,
        canvas: toga.Canvas,
        x: float | None = None,
        y: float | None = None,
        color: str = BLACK,
        fill_rule: FillRule = FillRule.NONZERO,
    ):
        super().__init__(canvas=canvas, x=x, y=y)
        self.color = color
        self.fill_rule = fill_rule

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.x}, y={self.y}, "
            f"color={self.color!r}, fill_rule={self.fill_rule})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.push_context(**kwargs)
        impl.begin_path(**kwargs)
        if self.x is not None and self.y is not None:
            impl.move_to(x=self.x, y=self.y, **kwargs)

        sub_kwargs = kwargs.copy()
        sub_kwargs.update(fill_color=self.color, fill_rule=self.fill_rule)
        for obj in self.drawing_objects:
            obj._draw(impl, **sub_kwargs)

        # Fill passes fill_rule to its children; but that is also a valid argument for
        # fill(), so if a fill context is a child of a fill context, there's an argument
        # collision. Duplicate the kwargs and explicitly overwrite to avoid the
        # collision.
        draw_kwargs = kwargs.copy()
        draw_kwargs.update(fill_rule=self.fill_rule)
        impl.fill(self.color, **draw_kwargs)

        impl.pop_context(**kwargs)

    @property
    def color(self) -> Color:
        """The fill color."""
        return self._color

    @color.setter
    def color(self, value: Color | str | None) -> None:
        if value is None:
            self._color = parse_color(BLACK)
        else:
            self._color = parse_color(value)


class StrokeContext(ClosedPathContext):
    """Construct a drawing context that will draw a stroke on all paths defined
    within the context.

    This is a context manager; it creates a new path, and moves to the start coordinate;
    when the context exits, the path is drawn with the stroke. For fine-grained control of
    a path, you can use :class:`~toga.widgets.canvas.Context.begin_path`,
    :class:`~toga.widgets.canvas.Context.move_to`,
    :class:`~toga.widgets.canvas.Context.close_path` and
    :class:`~toga.widgets.canvas.Context.stroke`.

    If both an x and y coordinate is provided, the drawing context will begin with
    a ``move_to`` operation in that context.

    You should not create a :class:`~toga.widgets.canvas.StrokeContext` context directly;
    instead, you should use the :meth:`~toga.widgets.canvas.Context.Stroke` method on
    an existing context.
    """

    def __init__(
        self,
        canvas: toga.Canvas,
        x: float | None = None,
        y: float | None = None,
        color: str | None = BLACK,
        line_width: float = 2.0,
        line_dash: list[float] | None = None,
    ):
        super().__init__(canvas=canvas, x=x, y=y)
        self.color = color
        self.line_width = line_width
        self.line_dash = line_dash

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.x}, y={self.y}, color={self.color!r}, "
            f"line_width={self.line_width}, line_dash={self.line_dash!r})"
        )

    def _draw(self, impl: Any, **kwargs: Any) -> None:
        impl.push_context(**kwargs)
        impl.begin_path(**kwargs)

        if self.x is not None and self.y is not None:
            impl.move_to(x=self.x, y=self.y, **kwargs)

        sub_kwargs = kwargs.copy()
        sub_kwargs["stroke_color"] = self.color
        sub_kwargs["line_width"] = self.line_width
        sub_kwargs["line_dash"] = self.line_dash
        for obj in self.drawing_objects:
            obj._draw(impl, **sub_kwargs)

        # Stroke passes line_width and line_dash to its children; but those two are also
        # valid arguments for stroke, so if a stroke context is a child of stroke
        # context, there's an argument collision. Duplicate the kwargs and explicitly
        # overwrite to avoid the collision
        draw_kwargs = kwargs.copy()
        draw_kwargs["line_width"] = self.line_width
        draw_kwargs["line_dash"] = self.line_dash
        impl.stroke(self.color, **draw_kwargs)

        impl.pop_context(**kwargs)

    @property
    def color(self) -> Color:
        """The color of the stroke."""
        return self._color

    @color.setter
    def color(self, value: object) -> None:
        if value is None:
            self._color = parse_color(BLACK)
        else:
            self._color = parse_color(value)


#######################################################################################
# Events
#######################################################################################


class OnTouchHandler(Protocol):
    def __call__(self, widget: Canvas, x: int, y: int, /, **kwargs: Any) -> object:
        """A handler that will be invoked when a :any:`Canvas` is touched with a finger
        or mouse.

        :param widget: The canvas that was touched.
        :param x: X coordinate, relative to the left edge of the canvas.
        :param y: Y coordinate, relative to the top edge of the canvas.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class OnResizeHandler(Protocol):
    def __call__(
        self, widget: Canvas, /, width: int, height: int, **kwargs: Any
    ) -> object:
        """A handler that will be invoked when a :any:`Canvas` is resized.

        :param widget: The canvas that was resized.
        :param width: The new width.
        :param height: The new height.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


#######################################################################################
# The Canvas Widget
#######################################################################################


class Canvas(Widget):
    _MIN_WIDTH = 0
    _MIN_HEIGHT = 0

    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        on_resize: OnResizeHandler | None = None,
        on_press: OnTouchHandler | None = None,
        on_activate: OnTouchHandler | None = None,
        on_release: OnTouchHandler | None = None,
        on_drag: OnTouchHandler | None = None,
        on_alt_press: OnTouchHandler | None = None,
        on_alt_release: OnTouchHandler | None = None,
        on_alt_drag: OnTouchHandler | None = None,
    ):
        """Create a new Canvas widget.

        Inherits from :class:`toga.Widget`.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param on_resize: Initial :any:`on_resize` handler.
        :param on_press: Initial :any:`on_press` handler.
        :param on_activate: Initial :any:`on_activate` handler.
        :param on_release: Initial :any:`on_release` handler.
        :param on_drag: Initial :any:`on_drag` handler.
        :param on_alt_press: Initial :any:`on_alt_press` handler.
        :param on_alt_release: Initial :any:`on_alt_release` handler.
        :param on_alt_drag: Initial :any:`on_alt_drag` handler.
        """

        super().__init__(id=id, style=style)

        self._context = Context(canvas=self)

        # Create a platform specific implementation of Canvas
        self._impl = self.factory.Canvas(interface=self)

        # Set all the properties
        self.on_resize = on_resize
        self.on_press = on_press
        self.on_activate = on_activate
        self.on_release = on_release
        self.on_drag = on_drag
        self.on_alt_press = on_alt_press
        self.on_alt_release = on_alt_release
        self.on_alt_drag = on_alt_drag

    @property
    def enabled(self) -> Literal[True]:
        """Is the widget currently enabled? i.e., can the user interact with the widget?
        ScrollContainer widgets cannot be disabled; this property will always return
        True; any attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value: object) -> None:
        pass

    def focus(self) -> None:
        """No-op; ScrollContainer cannot accept input focus."""
        pass

    @property
    def context(self) -> Context:
        """The root context for the canvas."""
        return self._context

    def redraw(self) -> None:
        """Redraw the Canvas.

        The Canvas will be automatically redrawn after adding or removing a drawing
        object, or when the Canvas resizes. However, when you modify the properties of a
        drawing object, you must call ``redraw`` manually.
        """
        self._impl.redraw()

    def Context(self) -> ContextManager[Context]:
        """Construct and yield a new sub-:class:`~toga.widgets.canvas.Context` within
        the root context of this Canvas.

        :yields: The new :class:`~toga.widgets.canvas.Context` object.
        """
        return self.context.Context()

    def ClosedPath(
        self,
        x: float | None = None,
        y: float | None = None,
    ) -> ContextManager[ClosedPathContext]:
        """Construct and yield a new :class:`~toga.widgets.canvas.ClosedPathContext` context in
        the root context of this canvas.

        :param x: The x coordinate of the path's starting point.
        :param y: The y coordinate of the path's starting point.
        :yields: The new :class:`~toga.widgets.canvas.ClosedPathContext` context object.
        """
        return self.context.ClosedPath(x, y)

    def Fill(
        self,
        x: float | None = None,
        y: float | None = None,
        color: Color | str | None = BLACK,
        fill_rule: FillRule = FillRule.NONZERO,
    ) -> ContextManager[FillContext]:
        """Construct and yield a new :class:`~toga.widgets.canvas.FillContext` in the
        root context of this canvas.

        A drawing operator that fills the path constructed in the context according to
        the current fill rule.

        If both an x and y coordinate is provided, the drawing context will begin with
        a ``move_to`` operation in that context.

        :param x: The x coordinate of the path's starting point.
        :param y: The y coordinate of the path's starting point.
        :param fill_rule: `nonzero` is the non-zero winding rule; `evenodd` is the
            even-odd winding rule.
        :param color: The fill color.
        :yields: The new :class:`~toga.widgets.canvas.FillContext` context object.
        """
        return self.context.Fill(x, y, color, fill_rule)

    def Stroke(
        self,
        x: float | None = None,
        y: float | None = None,
        color: Color | str | None = BLACK,
        line_width: float = 2.0,
        line_dash: list[float] | None = None,
    ) -> ContextManager[StrokeContext]:
        """Construct and yield a new :class:`~toga.widgets.canvas.StrokeContext` in the
        root context of this canvas.

        If both an x and y coordinate is provided, the drawing context will begin with
        a ``move_to`` operation in that context.

        :param x: The x coordinate of the path's starting point.
        :param y: The y coordinate of the path's starting point.
        :param color: The color for the stroke.
        :param line_width: The width of the stroke.
        :param line_dash: The dash pattern to follow when drawing the line. Default is a
            solid line.
        :yields: The new :class:`~toga.widgets.canvas.StrokeContext` context object.
        """
        return self.context.Stroke(x, y, color, line_width, line_dash)

    @property
    def on_resize(self) -> OnResizeHandler:
        """The handler to invoke when the canvas is resized."""
        return self._on_resize

    @on_resize.setter
    def on_resize(self, handler: OnResizeHandler) -> None:
        self._on_resize = wrapped_handler(self, handler)

    @property
    def on_press(self) -> OnTouchHandler:
        """The handler invoked when the canvas is pressed. When a mouse is being used,
        this press will be with the primary (usually the left) mouse button."""
        return self._on_press

    @on_press.setter
    def on_press(self, handler: OnTouchHandler) -> None:
        self._on_press = wrapped_handler(self, handler)

    @property
    def on_activate(self) -> OnTouchHandler:
        """The handler invoked when the canvas is pressed in a way indicating the
        pressed object should be activated. When a mouse is in use, this will usually be
        a double click with the primary (usually the left) mouse button.

        This event is not supported on Android or iOS."""
        return self._on_activate

    @on_activate.setter
    def on_activate(self, handler: OnTouchHandler) -> None:
        self._on_activate = wrapped_handler(self, handler)

    @property
    def on_release(self) -> OnTouchHandler:
        """The handler invoked when a press on the canvas ends."""
        return self._on_release

    @on_release.setter
    def on_release(self, handler: OnTouchHandler) -> None:
        self._on_release = wrapped_handler(self, handler)

    @property
    def on_drag(self) -> OnTouchHandler:
        """The handler invoked when the location of a press changes."""
        return self._on_drag

    @on_drag.setter
    def on_drag(self, handler: OnTouchHandler) -> None:
        self._on_drag = wrapped_handler(self, handler)

    @property
    def on_alt_press(self) -> OnTouchHandler:
        """The handler to invoke when the canvas is pressed in an alternate
        manner. This will usually correspond to a secondary (usually the right) mouse
        button press.

        This event is not supported on Android or iOS.
        """
        return self._on_alt_press

    @on_alt_press.setter
    def on_alt_press(self, handler: OnTouchHandler) -> None:
        self._on_alt_press = wrapped_handler(self, handler)

    @property
    def on_alt_release(self) -> OnTouchHandler:
        """The handler to invoke when an alternate press is released.

        This event is not supported on Android or iOS.
        """
        return self._on_alt_release

    @on_alt_release.setter
    def on_alt_release(self, handler: OnTouchHandler) -> None:
        self._on_alt_release = wrapped_handler(self, handler)

    @property
    def on_alt_drag(self) -> OnTouchHandler:
        """The handler to invoke when the location of an alternate press changes.

        This event is not supported on Android or iOS.
        """
        return self._on_alt_drag

    @on_alt_drag.setter
    def on_alt_drag(self, handler: OnTouchHandler) -> None:
        self._on_alt_drag = wrapped_handler(self, handler)

    ###########################################################################
    # Text measurement
    ###########################################################################

    def measure_text(
        self,
        text: str,
        font: Font | None = None,
        tight: None = None,  # DEPRECATED
    ) -> tuple[float, float]:
        """Measure the size at which :meth:`~.Context.write_text` would render some text.

        :param text: The text to measure. Newlines will cause line breaks, but long
            lines will not be wrapped.
        :param font: The font in which to draw the text. The default is the system font.
        :param tight: **DEPRECATED**: this argument has no effect.
        :returns: A tuple of ``(width, height)``.
        """
        if tight is not None:
            warnings.warn(
                "The `tight` argument on Canvas.measure_text() has been deprecated.",
                DeprecationWarning,
            )
        if font is None:
            font = Font(family=SYSTEM, size=SYSTEM_DEFAULT_FONT_SIZE)

        return self._impl.measure_text(str(text), font._impl)

    ###########################################################################
    # As image
    ###########################################################################

    def as_image(self, format: type[ImageT] = toga.Image) -> ImageT:
        """Render the canvas as an image.

        :param format: Format to provide. Defaults to :class:`~toga.images.Image`; also
            supports :any:`PIL.Image.Image` if Pillow is installed, as well as any image
            types defined by installed :doc:`image format plugins
            </reference/plugins/image_formats>`
        :returns: The canvas as an image of the specified type.
        """
        return toga.Image(self._impl.get_image_data()).as_format(format)

    ###########################################################################
    # 2023-07 Backwards compatibility
    ###########################################################################

    def new_path(self):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.begin_path` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.begin_path()",
            DeprecationWarning,
        )
        return self.context.begin_path()

    def move_to(self, x, y):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.move_to` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.move_to()",
            DeprecationWarning,
        )
        return self.context.move_to(x, y)

    def line_to(self, x, y):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.line_to` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.line_to()",
            DeprecationWarning,
        )
        return self.context.line_to(x, y)

    def bezier_curve_to(self, cp1x, cp1y, cp2x, cp2y, x, y):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.bezier_curve_to` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.bezier_curve_to()",
            DeprecationWarning,
        )
        return self.context.bezier_curve_to(cp1x, cp1y, cp2x, cp2y, x, y)

    def quadratic_curve_to(self, cpx: float, cpy: float, x: float, y: float):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.quadratic_curve_to`
        on :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.quadratic_curve_to()",
            DeprecationWarning,
        )
        return self.context.quadratic_curve_to(cpx, cpy, x, y)

    def arc(self, x, y, radius, startangle=0.0, endangle=2 * pi, anticlockwise=False):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.arc` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.arc()",
            DeprecationWarning,
        )
        return self.context.arc(x, y, radius, startangle, endangle, anticlockwise)

    def ellipse(
        self,
        x: float,
        y: float,
        radiusx: float,
        radiusy: float,
        rotation: float = 0.0,
        startangle: float = 0.0,
        endangle: float = 2 * pi,
        anticlockwise: bool = False,
    ):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.ellipse` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.ellipse()",
            DeprecationWarning,
        )
        return self.context.ellipse(
            x,
            y,
            radiusx,
            radiusy,
            rotation,
            startangle,
            endangle,
            anticlockwise,
        )

    def rect(self, x: float, y: float, width: float, height: float):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.rect` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.rect()",
            DeprecationWarning,
        )
        return self.context.rect(x, y, width, height)

    def write_text(self, text: str, x=0, y=0, font=None):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.write_text` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.write_text()",
            DeprecationWarning,
        )
        return self.context.write_text(text, x, y, font)

    def rotate(self, radians: float):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.rotate` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.rotate()",
            DeprecationWarning,
        )
        return self.context.rotate(radians)

    def scale(self, sx: float, sy: float):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.scale` on :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.scale()",
            DeprecationWarning,
        )
        return self.context.scale(sx, sy)

    def translate(self, tx: float, ty: float):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.translate` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.translate()",
            DeprecationWarning,
        )
        return self.context.translate(tx, ty)

    def reset_transform(self):
        """**DEPRECATED** - Use :meth:`~toga.widgets.canvas.Context.reset_transform` on
        :attr:`context`"""
        warnings.warn(
            "Direct canvas operations have been deprecated; use context.reset_transform()",
            DeprecationWarning,
        )
        return self.context.reset_transform()

    def closed_path(self, x, y):
        """**DEPRECATED** - use :meth:`~toga.Canvas.ClosedPath`"""
        warnings.warn(
            "Canvas.closed_path() has been renamed Canvas.ClosedPath()",
            DeprecationWarning,
        )
        return self.ClosedPath(x, y)

    def fill(
        self,
        color: Color | str | None = BLACK,
        fill_rule: FillRule = FillRule.NONZERO,
        preserve=None,  # DEPRECATED
    ):
        """**DEPRECATED** - use :meth:`~toga.Canvas.Fill`"""
        warnings.warn(
            "Canvas.fill() has been renamed Canvas.Fill()",
            DeprecationWarning,
        )
        if preserve is not None:
            warnings.warn(
                "The `preserve` argument on fill() has been deprecated.",
                DeprecationWarning,
            )
        return self.Fill(color=color, fill_rule=fill_rule)

    def stroke(
        self,
        color: Color | str | None = BLACK,
        line_width: float = 2.0,
        line_dash: list[float] | None = None,
    ):
        """**DEPRECATED** - use :meth:`~toga.Canvas.Stroke`"""
        warnings.warn(
            "Canvas.stroke() has been renamed Canvas.Stroke().",
            DeprecationWarning,
        )
        return self.Stroke(color=color, line_width=line_width, line_dash=line_dash)


def sweepangle(startangle: float, endangle: float, anticlockwise: bool) -> float:
    """Returns an arc length in the range [-2 * pi, 2 * pi], where positive numbers are
    clockwise. Based on the "ellipse method steps" in the HTML spec."""

    if anticlockwise:
        if endangle - startangle <= -2 * pi:
            return -2 * pi
    else:
        if endangle - startangle >= 2 * pi:
            return 2 * pi

    startangle %= 2 * pi
    endangle %= 2 * pi
    sweepangle = endangle - startangle
    if anticlockwise:
        if sweepangle > 0:
            sweepangle -= 2 * pi
    else:
        if sweepangle < 0:
            sweepangle += 2 * pi

    return sweepangle


# Based on https://stackoverflow.com/a/30279817
def arc_to_bezier(sweepangle: float) -> list[tuple[float, float]]:
    """Approximates an arc of a unit circle as a sequence of Bezier segments.

    :param sweepangle: Length of the arc in radians, where positive numbers are
        clockwise.
    :returns: [(1, 0), (cp1x, cp1y), (cp2x, cp2y), (x, y), ...], where each group of 3
        points has the same meaning as in the bezier_curve_to method, and there are
        between 1 and 4 groups."""

    matrices = [
        [1, 0, 0, 1],  # 0 degrees
        [0, -1, 1, 0],  # 90
        [-1, 0, 0, -1],  # 180
        [0, 1, -1, 0],  # 270
    ]

    if sweepangle < 0:  # Anticlockwise
        sweepangle *= -1
        for matrix in matrices:
            matrix[2] *= -1
            matrix[3] *= -1

    result = [(1.0, 0.0)]
    for matrix in matrices:
        if sweepangle < 0:
            break

        phi = min(sweepangle, pi / 2)
        k = 4 / 3 * tan(phi / 4)
        result += [
            transform(x, y, matrix)
            for x, y in [
                (1, k),
                (cos(phi) + k * sin(phi), sin(phi) - k * cos(phi)),
                (cos(phi), sin(phi)),
            ]
        ]

        sweepangle -= pi / 2

    return result


def transform(x: float, y: float, matrix: list[int]) -> tuple[float, float]:
    return (
        x * matrix[0] + y * matrix[1],
        x * matrix[2] + y * matrix[3],
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Protocol, SupportsInt

from toga.handlers import wrapped_handler
from toga.types import Position

from .base import StyleT, Widget

if TYPE_CHECKING:
    from toga.types import PositionT


class OnScrollHandler(Protocol):
    def __call__(self, widget: ScrollContainer, /, **kwargs: Any) -> object:
        """A handler to invoke when the container is scrolled.

        :param widget: The ScrollContainer that was scrolled.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class ScrollContainer(Widget):
    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        horizontal: bool = True,
        vertical: bool = True,
        on_scroll: OnScrollHandler | None = None,
        content: Widget | None = None,
    ):
        """Create a new Scroll Container.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param horizontal: Should horizontal scrolling be permitted?
        :param vertical: Should horizontal scrolling be permitted?
        :param on_scroll: Initial :any:`on_scroll` handler.
        :param content: The content to display in the scroll window.
        """
        super().__init__(id=id, style=style)

        self._content: Widget | None = None
        self.on_scroll = None

        # Create a platform specific implementation of a Scroll Container
        self._impl = self.factory.ScrollContainer(interface=self)

        # Set all attributes
        self.vertical = vertical
        self.horizontal = horizontal
        self.content = content
        self.on_scroll = on_scroll

    @Widget.app.setter
    def app(self, app) -> None:
        # Invoke the superclass property setter
        Widget.app.fset(self, app)

        # Also assign the app to the content in the container
        if self._content:
            self._content.app = app

    @Widget.window.setter
    def window(self, window) -> None:
        # Invoke the superclass property setter
        Widget.window.fset(self, window)

        # Also assign the window to the content in the container
        if self._content:
            self._content.window = window

    @property
    def enabled(self) -> Literal[True]:
        """Is the widget currently enabled? i.e., can the user interact with the widget?

        ScrollContainer widgets cannot be disabled; this property will always return
        True; any attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value: object) -> None:
        pass

    def focus(self) -> None:
        """No-op; ScrollContainer cannot accept input focus."""
        pass

    @property
    def content(self) -> Widget | None:
        """The root content widget displayed inside the scroll container."""
        return self._content

    @content.setter
    def content(self, widget: Widget | None) -> None:
        if self._content:
            # Clear the window before the app so that registry entries can be cleared
            self._content.window = None
            self._content.app = None

        if widget:
            widget.app = self.app
            widget.window = self.window
            self._impl.set_content(widget._impl)
        else:
            self._impl.set_content(None)

        self._content = widget
        if widget:
            widget.refresh()

    @property
    def vertical(self) -> bool:
        """Is vertical scrolling enabled?"""
        return self._impl.get_vertical()

    @vertical.setter
    def vertical(self, value: object) -> None:
        self._impl.set_vertical(bool(value))
        if self._content:
            self._content.refresh()

    @property
    def horizontal(self) -> bool:
        """Is horizontal scrolling enabled?"""
        return self._impl.get_horizontal()

    @horizontal.setter
    def horizontal(self, value: object) -> None:
        self._impl.set_horizontal(bool(value))
        if self._content:
            self._content.refresh()

    @property
    def on_scroll(self) -> OnScrollHandler:
        """Handler to invoke when the user moves a scroll bar."""
        return self._on_scroll

    @on_scroll.setter
    def on_scroll(self, on_scroll: OnScrollHandler) -> None:
        self._on_scroll = wrapped_handler(self, on_scroll)

    @property
    def max_horizontal_position(self) -> int:
        """The maximum horizontal scroll position (read-only)."""
        if not self.horizontal:
            return 0
        else:
            return self._impl.get_max_horizontal_position()

    @property
    def horizontal_position(self) -> int:
        """The current horizontal scroll position.

        If the value provided is negative, or greater than the maximum horizontal
        position, the value will be clipped to the valid range.

        :returns: The current horizontal scroll position.
        :raises ValueError: If an attempt is made to change the horizontal position when
            horizontal scrolling is disabled.
        """
        return self._impl.get_horizontal_position()

    @horizontal_position.setter
    def horizontal_position(self, horizontal_position: SupportsInt) -> None:
        if not self.horizontal:
            raise ValueError(
                "Cannot set horizontal position when horizontal scrolling is not enabled."
            )

        self.position = (horizontal_position, self._impl.get_vertical_position())

    @property
    def max_vertical_position(self) -> int:
        """The maximum vertical scroll position (read-only)."""
        if not self.vertical:
            return 0
        else:
            return self._impl.get_max_vertical_position()

    @property
    def vertical_position(self) -> int:
        """The current vertical scroll position.

        If the value provided is negative, or greater than the maximum
        vertical position, the value will be clipped to the valid range.

        :returns: The current vertical scroll position.
        :raises ValueError: If an attempt is made to change the vertical position
            when vertical scrolling is disabled.
        """
        return self._impl.get_vertical_position()

    @vertical_position.setter
    def vertical_position(self, vertical_position: SupportsInt) -> None:
        if not self.vertical:
            raise ValueError(
                "Cannot set vertical position when vertical scrolling is not enabled."
            )

        self.position = (self._impl.get_horizontal_position(), vertical_position)

    # This combined property is necessary because on some platforms (e.g. iOS), setting
    # the horizontal and vertical position separately would cause the horizontal and
    # vertical movement to appear as two separate animations.
    @property
    def position(self) -> Position:
        """The current scroll position.

        If the value provided for either axis is negative, or greater than the maximum
        position in that axis, the value will be clipped to the valid range.

        If scrolling is disabled in either axis, the value provided for that axis will
        be ignored.
        """
        return Position(self.horizontal_position, self.vertical_position)

    @position.setter
    def position(self, position: PositionT):
        horizontal_position, vertical_position = map(int, position)
        if self.horizontal:
            if horizontal_position < 0:
                horizontal_position = 0
            else:
                max_value = self.max_horizontal_position
                if horizontal_position > max_value:
                    horizontal_position = max_value
        else:
            horizontal_position = self.horizontal_position

        if self.vertical:
            if vertical_position < 0:
                vertical_position = 0
            else:
                max_value = self.max_vertical_position
                if vertical_position > max_value:
                    vertical_position = max_value
        else:
            vertical_position = self.vertical_position

        self._impl.set_position(horizontal_position, vertical_position)



import pytest

import toga
from toga.style import Pack
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
    assert_attribute_not_set,
    attribute_value,
)


# Create the simplest possible widget with a concrete implementation that will
# allow children
class ExampleWidget(toga.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._impl = self.factory.Widget(self)
        self._children = []


# Create the simplest possible widget with a concrete implementation that cannot
# have children.
class ExampleLeafWidget(toga.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._impl = self.factory.Widget(self)


@pytest.fixture
def widget(app):
    # App fixture is needed to ensure a fresh widget registry is created
    return ExampleWidget(id="widget_id", style=Pack(padding=666))


def test_simple_widget():
    """A simple widget can be created."""
    widget = ExampleWidget()

    # Round trip the impl/interface
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create Widget")

    # Base properties of the widget have been set
    assert widget.id == str(id(widget))
    assert isinstance(widget.style, Pack)
    assert widget.style.padding == (0, 0, 0, 0)


def test_widget_created(widget):
    """A widget can be created with arguments."""
    # Round trip the impl/interface
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create Widget")

    # Base properties of the widget have been set
    assert widget.enabled
    assert widget.id == "widget_id"
    assert isinstance(widget.style, Pack)
    assert widget.style.padding == (666, 666, 666, 666)


def test_add_child_to_leaf():
    """A child cannot be added to a leaf node."""
    leaf = ExampleLeafWidget()

    # Widget doesn't have an app or window
    assert leaf.app is None
    assert leaf.window is None

    # Leaf nodes report an empty child list
    assert leaf.children == []

    # Create a child widget
    child = ExampleLeafWidget()

    # Add the child.
    with pytest.raises(ValueError, match=r"ExampleLeafWidget cannot have children"):
        leaf.add(child)


def test_add_child_without_app(widget):
    """A child can be added to a node when there's no underlying app."""
    # Widget doesn't have an app or window
    assert widget.app is None
    assert widget.window is None

    # Child list is empty
    assert widget.children == []

    # Create a child widget
    child = ExampleLeafWidget()

    # Add the child.
    widget.add(child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # Child has inherited parent's app/window details
    assert child.app is None
    assert child.window is None

    # The impl's add_child has been invoked
    assert_action_performed_with(widget, "add child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")


def test_add_child(app, widget):
    """A child can be added to a node when there's an app & window."""
    # Set the app and window for the widget.
    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    # Widget has an app and window
    assert widget.app == app
    assert widget.window == window

    # Widget is registered with app and window
    assert widget.id in app.widgets
    assert widget.id in window.widgets

    # Child list is empty
    assert widget.children == []

    # Create a child widget
    child = ExampleLeafWidget(id="child_id")

    # App's widget index only contains the parent
    assert app.widgets["widget_id"] == widget
    assert "child_id" not in app.widgets

    # Add the child.
    widget.add(child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # Child has inherited parent's app/window details
    assert child.app == app
    assert child.window == window

    # The impl's add_child has been invoked
    assert_action_performed_with(widget, "add child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The window's content gets a refresh notification
    assert_action_performed_with(window.content, "refresh")

    # App's widget index has been updated
    assert len(app.widgets) == 2
    assert app.widgets["widget_id"] == widget
    assert app.widgets["child_id"] == child


def test_child_square_bracket_access(widget):
    """Assign and get individual children via square bracket notation."""
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")

    # Add child to widget
    widget.add(child1)

    # Verify child has been added
    assert widget.children == [child1]

    # Assign child via __setitem__
    widget.children[0] = child2

    # Verify new child is in children via __getitem__
    assert widget.children[0] == child2


def test_get_index_of_children(widget):
    """Populate a widget with children and check the index method."""
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")

    # Add only two children to widget
    widget.add(child1)
    widget.add(child2)

    # Verify children have been added
    assert widget.children == [child1, child2]

    # Get indices of valid children
    assert widget.index(child1) == 0
    assert widget.index(child2) == 1

    # Raise error when child not found in children
    with pytest.raises(ValueError, match=r"ExampleLeafWidget not found"):
        widget.index(child3)


def test_replace_child_with_non_child(widget):
    """Remove a child and add new widget in its place."""
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")

    # Add only two children to widget
    widget.add(child1)
    widget.add(child2)

    # Verify children have been added
    assert widget.children == [child1, child2]

    widget.replace(child1, child3)

    # Verify child1 has been removed and child3 has been added
    assert widget.children == [child3, child2]


def test_add_multiple_children(app, widget):
    """Multiple children can be added in one call."""
    # Set the app and window for the widget.
    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    # Widget has an app and window
    assert widget.app == app
    assert widget.window == window

    # Child list is empty
    assert widget.children == []

    # Create 3 child widgets
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")

    # App's widget index only contains the parent
    assert dict(app.widgets) == {"widget_id": widget}

    # Add the children.
    widget.add(child1, child2, child3)

    # Widget knows about the child and vice versa
    assert widget.children == [child1, child2, child3]
    assert child1.parent == widget
    assert child2.parent == widget
    assert child3.parent == widget

    # Children has inherited parent's app/window details
    assert child1.app == app
    assert child1.window == window

    assert child2.app == app
    assert child2.window == window

    assert child3.app == app
    assert child3.window == window

    # The impl's add_child has been invoked 3 time
    assert_action_performed_with(widget, "add child", child=child1._impl)
    assert_action_performed_with(widget, "add child", child=child2._impl)
    assert_action_performed_with(widget, "add child", child=child3._impl)

    # The widget's layout has been refreshed
    # There will be multiple refresh calls
    assert_action_performed_with(widget, "refresh")

    # The window's content gets a refresh notification
    assert_action_performed_with(window.content, "refresh")

    # App's widget index has been updated
    assert dict(app.widgets) == {
        "widget_id": widget,
        "child1_id": child1,
        "child2_id": child2,
        "child3_id": child3,
    }

    # Window's widget index has been updated
    assert dict(window.widgets) == {
        "widget_id": widget,
        "child1_id": child1,
        "child2_id": child2,
        "child3_id": child3,
    }


def test_reparent_child(widget):
    """A widget can be reparented."""
    # Create a second parent widget, and add a child to it
    other = ExampleWidget(id="other")
    child = ExampleLeafWidget(id="child_id")
    other.add(child)

    assert other.children == [child]
    assert child.parent == other

    # Add the child to the widget
    widget.add(child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # Original parent as lost the child
    assert other.children == []

    # The impl's add_child has been invoked
    assert_action_performed_with(widget, "add child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The layout of the old parent has been refreshed
    assert_action_performed_with(other, "refresh")


def test_reparent_child_to_self(widget):
    """Reparenting a widget to the same parent is a no-op."""
    # Add a child to the widget
    child = ExampleLeafWidget(id="child_id")
    widget.add(child)

    assert widget.children == [child]
    assert child.parent == widget

    # Reset the event log so all previous add events are lost
    EventLog.reset()

    # Add the child to the widget again
    widget.add(child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # The impl's add_child has *not* been invoked,
    # as the widget was already a child
    assert_action_not_performed(widget, "add child")

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")


def test_insert_child_into_leaf():
    """A child cannot be inserted into a leaf node."""
    leaf = ExampleLeafWidget()

    # Widget doesn't have an app or window
    assert leaf.app is None
    assert leaf.window is None

    # Leaf nodes report an empty child list
    assert leaf.children == []

    # Create a child widget
    child = ExampleLeafWidget()

    # insert the child.
    with pytest.raises(ValueError, match=r"ExampleLeafWidget cannot have children"):
        leaf.insert(0, child)


def test_insert_child_without_app(widget):
    """A child can be inserted into a node when there's no underlying app."""
    # Widget doesn't have an app or window
    assert widget.app is None
    assert widget.window is None

    # Child list is empty
    assert widget.children == []

    # Create a child widget
    child = ExampleLeafWidget()

    # insert the child.
    widget.insert(0, child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # Child has inherited parent's app/window details
    assert child.app is None
    assert child.window is None

    # The impl's insert_child has been invoked
    assert_action_performed_with(widget, "insert child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")


def test_insert_child(app, widget):
    """A child can be inserted into a node when there's an app & window."""
    # Set the app and window for the widget.
    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    # Widget has an app and window
    assert widget.app == app
    assert widget.window == window

    # Child list is empty
    assert widget.children == []

    # Create a child widget
    child = ExampleLeafWidget(id="child_id")

    # App's widget index only contains the parent
    assert app.widgets["widget_id"] == widget
    assert "child_id" not in app.widgets

    # insert the child.
    widget.insert(0, child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # Child has inherited parent's app/window details
    assert child.app == app
    assert child.window == window

    # The impl's insert_child has been invoked
    assert_action_performed_with(widget, "insert child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The window's content gets a refresh notification
    assert_action_performed_with(window.content, "refresh")

    # App's widget index has been updated
    assert dict(app.widgets) == {
        "widget_id": widget,
        "child_id": child,
    }

    # Window's widget index has been updated
    assert dict(window.widgets) == {
        "widget_id": widget,
        "child_id": child,
    }


def test_insert_position(app, widget):
    """Insert can put a child into a specific position."""
    # Set the app and window for the widget.
    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    # Widget has an app and window
    assert widget.app == app
    assert widget.window == window

    # Child list is empty
    assert widget.children == []

    # Create 3 child widgets
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")

    # App's widget index only contains the parent
    assert dict(app.widgets) == {"widget_id": widget}

    # Windows's widget index only contains the parent
    assert dict(window.widgets) == {"widget_id": widget}

    # insert the children.
    widget.insert(0, child1)
    widget.insert(0, child2)
    widget.insert(1, child3)

    # Widget knows about the child and vice versa
    assert widget.children == [child2, child3, child1]
    assert child1.parent == widget
    assert child2.parent == widget
    assert child3.parent == widget

    # Children has inherited parent's app/window details
    assert child1.app == app
    assert child1.window == window

    assert child2.app == app
    assert child2.window == window

    assert child3.app == app
    assert child3.window == window

    # The impl's insert_child has been invoked 3 time
    assert_action_performed_with(widget, "insert child", child=child1._impl)
    assert_action_performed_with(widget, "insert child", child=child2._impl)
    assert_action_performed_with(widget, "insert child", child=child3._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The window's content gets a refresh notification
    assert_action_performed_with(window.content, "refresh")

    # App's widget index has been updated
    assert dict(app.widgets) == {
        "widget_id": widget,
        "child1_id": child1,
        "child2_id": child2,
        "child3_id": child3,
    }

    # Window's widget index has been updated
    assert dict(window.widgets) == {
        "widget_id": widget,
        "child1_id": child1,
        "child2_id": child2,
        "child3_id": child3,
    }


def test_insert_bad_position(app, widget):
    """If the position is invalid, an error is raised."""
    # Set the app and window for the widget.
    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    # Widget has an app and window
    assert widget.app == app
    assert widget.window == window

    # Child list is empty
    assert widget.children == []

    # Create a child widget
    child = ExampleLeafWidget(id="child_id")

    # App's widget index only contains the parent
    assert dict(app.widgets) == {"widget_id": widget}

    # Window's widget index only contains the parent
    assert dict(window.widgets) == {"widget_id": widget}

    # Insert the child at a position greater than the length of the list.
    # Widget will be added to the end of the list.
    widget.insert(37, child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # Child has inherited parent's app/window details
    assert child.app == app
    assert child.window == window

    # The impl's insert_child has been invoked
    assert_action_performed_with(widget, "insert child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The window's content gets a refresh notification
    assert_action_performed_with(window.content, "refresh")

    # App's widget index has been updated
    assert dict(app.widgets) == {
        "widget_id": widget,
        "child_id": child,
    }

    # Window's widget index has been updated
    assert dict(window.widgets) == {
        "widget_id": widget,
        "child_id": child,
    }


def test_insert_reparent_child(widget):
    """A widget can be reparented by insertion."""
    # Create a second parent widget, and add a child to it
    other = ExampleWidget(id="other")
    child = ExampleLeafWidget(id="child_id")
    other.add(child)

    assert other.children == [child]
    assert child.parent == other

    # insert the child to the widget
    widget.insert(0, child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # Original parent as lost the child
    assert other.children == []

    # The impl's insert_child has been invoked
    assert_action_performed_with(widget, "insert child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The original parent's layout has been refreshed
    assert_action_performed_with(other, "refresh")


def test_insert_reparent_child_to_self(widget):
    """Reparenting a widget to the same parent by insertion is a no-op."""
    # Add a child to the widget
    child = ExampleLeafWidget(id="child_id")
    widget.add(child)

    assert widget.children == [child]
    assert child.parent == widget

    # Reset the event log so all previous insert events are lost
    EventLog.reset()

    # insert the child to the widget again
    widget.insert(0, child)

    # Widget knows about the child and vice versa
    assert widget.children == [child]
    assert child.parent == widget

    # The impl's insert_child has *not* been invoked,
    # as the widget was already a child
    assert_action_not_performed(widget, "insert child")

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")


def test_remove_child_from_leaf():
    """A child cannot be removed from a leaf node."""
    leaf = ExampleLeafWidget()

    # Widget doesn't have an app or window
    assert leaf.app is None
    assert leaf.window is None

    # Leaf nodes report an empty child list
    assert leaf.children == []

    # Create a child widget
    child = ExampleLeafWidget()

    # Remove the child.
    with pytest.raises(ValueError, match=r"ExampleLeafWidget cannot have children"):
        leaf.remove(child)


def test_remove_child_without_app(widget):
    """A child without an app or window can be removed from a widget."""
    # Add a child to the widget
    child = ExampleLeafWidget(id="child_id")
    widget.add(child)

    assert widget.children == [child]
    assert child.parent == widget
    assert child.app is None
    assert child.window is None

    # Remove the child
    widget.remove(child)

    # Parent doesn't know about the child, and vice versa
    assert widget.children == []
    assert child.parent is None

    # App and window are still None
    assert child.app is None
    assert child.window is None

    # The impl's remove_child has been invoked
    assert_action_performed_with(widget, "remove child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")


def test_remove_child(app, widget):
    """A child associated with an app & window can be removed from a widget."""
    # Add a child to the widget
    child = ExampleLeafWidget(id="child_id")
    widget.add(child)

    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    assert widget.children == [child]
    assert child.parent == widget
    assert child.app == app
    assert child.window == window
    assert dict(app.widgets) == {"widget_id": widget, "child_id": child}
    assert dict(window.widgets) == {"widget_id": widget, "child_id": child}

    # Remove the child
    widget.remove(child)

    # Parent doesn't know about the child, and vice versa
    assert widget.children == []
    assert child.parent is None

    # app and window have been reset.
    assert child.app is None
    assert child.window is None

    # child widget no longer exists in the app or widgets registries.
    assert dict(app.widgets) == {"widget_id": widget}
    assert dict(window.widgets) == {"widget_id": widget}

    # The impl's remove_child has been invoked
    assert_action_performed_with(widget, "remove child", child=child._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The window's content gets a refresh notification
    assert_action_performed_with(window.content, "refresh")

    # App's widget index does not contain the widget
    assert "child_id" not in app.widgets


def test_remove_multiple_children(app, widget):
    """Multiple children can be removed from a widget."""
    # Add children to the widget
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")
    widget.add(child1, child2, child3)

    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    assert widget.children == [child1, child2, child3]
    for child in widget.children:
        assert child.parent == widget
        assert child.app == app
        assert child.window == window
        assert app.widgets[child.id] == child
        assert window.widgets[child.id] == child

    # Remove 2 children
    widget.remove(child1, child3)

    # Parent doesn't know about the removed children, and vice versa
    assert widget.children == [child2]
    assert child1.parent is None
    assert child2.parent == widget
    assert child3.parent is None

    # App and window have been reset on the removed widgets
    assert child1.app is None
    assert child1.window is None

    assert child2.app == app
    assert child2.window == window

    assert child3.app is None
    assert child3.window is None

    # The impl's remove_child has been invoked twice
    assert_action_performed_with(widget, "remove child", child=child1._impl)
    assert_action_performed_with(widget, "remove child", child=child3._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The window's content gets a refresh notification
    assert_action_performed_with(window.content, "refresh")

    # App's widget index does not contain the widget
    assert "child1_id" not in app.widgets
    assert "child3_id" not in app.widgets

    # Windows's widget index does not contain the widget
    assert "child1_id" not in window.widgets
    assert "child3_id" not in window.widgets


def test_clear_all_children(app, widget):
    """All children can be simultaneously removed from a widget."""
    # Add children to the widget
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")
    widget.add(child1, child2, child3)

    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    assert widget.children == [child1, child2, child3]
    for child in widget.children:
        assert child.parent == widget
        assert child.app == app
        assert child.window == window
        assert app.widgets[child.id] == child
        assert window.widgets[child.id] == child

    # Clear children
    widget.clear()

    # Parent doesn't know about the removed children, and vice versa
    assert widget.children == []
    assert child1.parent is None
    assert child2.parent is None
    assert child3.parent is None

    # App and window have been reset on the removed widgets
    assert child1.app is None
    assert child1.window is None

    assert child2.app is None
    assert child2.window is None

    assert child3.app is None
    assert child3.window is None

    # The impl's remove_child has been invoked thrice
    assert_action_performed_with(widget, "remove child", child=child1._impl)
    assert_action_performed_with(widget, "remove child", child=child2._impl)
    assert_action_performed_with(widget, "remove child", child=child3._impl)

    # The widget's layout has been refreshed
    assert_action_performed_with(widget, "refresh")

    # The window's content gets a refresh notification
    assert_action_performed_with(window.content, "refresh")

    # App's widget index does not contain the widget
    assert "child1_id" not in app.widgets
    assert "child2_id" not in app.widgets
    assert "child3_id" not in app.widgets

    # Window's widget index does not contain the widget
    assert "child1_id" not in window.widgets
    assert "child2_id" not in window.widgets
    assert "child3_id" not in window.widgets


def test_clear_no_children(app, widget):
    """No changes are made (no-op) if widget has no children."""
    window = toga.Window()
    window.content = widget
    # Clear the event log
    EventLog.reset()

    assert widget.children == []

    # Clear children
    widget.clear()

    # Parent doesn't have any children still
    assert widget.children == []

    # The widget's layout has *not* been refreshed
    assert_action_not_performed(widget, "refresh")

    # The window's content doesn't get a refresh notification
    assert_action_not_performed(window.content, "refresh")


def test_clear_leaf(app):
    """`clear` cannot be called on a leaf node."""
    leaf = ExampleLeafWidget()
    window = toga.Window()
    window.content = leaf
    # Clear the event log
    EventLog.reset()

    assert leaf.children == []

    # Clear children
    with pytest.raises(ValueError, match=r"ExampleLeafWidget cannot have children"):
        leaf.clear()

    # Parent doesn't have any children still
    assert leaf.children == []

    # The widget's layout has *not* been refreshed
    assert_action_not_performed(leaf, "refresh")

    # The window's content doesn't get a refresh notification
    assert_action_not_performed(window.content, "refresh")


def test_remove_from_non_parent(widget):
    """Trying to remove a child from a widget other than it's parent is a no-op."""
    # Create a second parent widget, and add a child to it
    other = ExampleWidget(id="other")
    child = ExampleLeafWidget(id="child_id")
    other.add(child)

    assert widget.children == []
    assert other.children == [child]
    assert child.parent == other

    # Remove the child from *widget*, which is not the parent
    widget.remove(child)

    # Nothing has changed.
    assert widget.children == []
    assert other.children == [child]
    assert child.parent == other

    # The impl's remove_child has been invoked
    assert_action_not_performed(widget, "remove child")

    # The widget's layout has *not* been refreshed
    assert_action_not_performed(widget, "refresh")


def test_set_app(app, widget):
    """A widget can be assigned to an app."""
    assert len(app.widgets) == 0

    # Assign the widget to an app
    widget.app = app

    # The app has been assigned
    assert widget.app == app

    # The widget doesn't appear in the index, as it's not on a window
    assert len(app.widgets) == 0
    assert "widget_id" not in app.widgets

    # The impl has had its app property set.
    assert attribute_value(widget, "app") == app

    # Assign the widget to a window
    window = toga.Window()
    window.content = widget

    # The widget is now in the app index.
    assert len(app.widgets) == 1
    assert app.widgets["widget_id"] == widget


def test_set_app_with_children(app, widget):
    """If a widget has children, the children get the app assignment."""
    # Add children to the widget
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")
    widget.add(child1, child2, child3)

    assert len(app.widgets) == 0

    # Assign the widget as window content
    window = toga.Window()
    window.content = widget

    # The app has been assigned
    assert widget.app == app

    # The children also have the app assigned
    assert child1.app == app
    assert child2.app == app
    assert child3.app == app

    # The widget index has been updated
    assert len(app.widgets) == 4
    assert app.widgets["widget_id"] == widget
    assert app.widgets["child1_id"] == child1
    assert app.widgets["child2_id"] == child2
    assert app.widgets["child3_id"] == child3

    # The impl of widget and children have had their app property set.
    assert attribute_value(widget, "app") == app
    assert attribute_value(child1, "app") == app
    assert attribute_value(child2, "app") == app
    assert attribute_value(child3, "app") == app


def test_set_same_app(app, widget):
    """A widget can be re-assigned to the same app."""
    assert len(app.widgets) == 0

    # Assign the widget to an app
    widget.app = app

    # Reset the event log so we know the new events
    EventLog.reset()

    # Assign the widget to the same app
    widget.app = app

    # The impl has not had its app property set as a result of the update
    assert_attribute_not_set(widget, "app")


def test_reset_app(app, widget):
    """A widget can be re-assigned to no app."""
    assert len(app.widgets) == 0

    # Assign the widget to an app
    widget.app = app

    # Reset the event log so we know the new events
    EventLog.reset()

    # Clear the app assignment
    widget.app = None

    # The app has been assigned
    assert widget.app is None

    # The widget index has been updated
    assert len(app.widgets) == 0

    # The impl has had its app property set.
    assert attribute_value(widget, "app") is None


def test_set_new_app(app, widget):
    """A widget can be assigned to a different app."""
    # Assign the widget to an app. It won't appear in the registry, as
    # it hasn't been assigned to a window
    widget.app = app
    assert len(app.widgets) == 0

    # Reset the event log so we know the new events
    EventLog.reset()

    # Create a new app
    new_app = toga.App("Test App", "org.beeware.toga.test-app")
    assert len(new_app.widgets) == 0

    # Assign the widget to the same app
    widget.app = new_app

    # The widget has been assigned to the new app
    assert widget.app == new_app

    # The widget still doesn't appear in a widget index.
    assert len(app.widgets) == 0
    assert "widget_id" not in app.widgets
    assert len(new_app.widgets) == 0
    assert "widget_id" not in new_app.widgets

    # The impl has had its app property set.
    assert attribute_value(widget, "app") == new_app


def test_set_window(widget):
    """A widget can be assigned to a window."""
    window = toga.Window()
    assert len(window.widgets) == 0
    assert widget.window is None

    # Assign the widget to a window
    widget.window = window

    # Window has been assigned
    assert widget.window == window

    # Window Widget registry has been updated
    assert len(window.widgets) == 1
    assert window.widgets["widget_id"] == widget


def test_set_window_with_children(app, widget):
    """A widget can be assigned to a window."""
    # Add children to the widget
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")
    widget.add(child1, child2, child3)

    window = toga.Window()
    assert len(app.widgets) == 0
    assert len(window.widgets) == 0
    assert widget.window is None
    assert child1.window is None
    assert child2.window is None
    assert child3.window is None

    # Assign the widget to a window
    widget.window = window

    # Window has been assigned
    assert widget.window == window
    assert child1.window == window
    assert child2.window == window
    assert child3.window == window

    # Widget registry has been updated
    assert len(app.widgets) == 4
    assert len(window.widgets) == 4
    assert window.widgets["widget_id"] == widget
    assert window.widgets["child1_id"] == child1
    assert window.widgets["child2_id"] == child2
    assert window.widgets["child3_id"] == child3


def test_reset_window(widget):
    """A widget can be assigned to a different window."""
    window = toga.Window()
    assert len(window.widgets) == 0
    assert widget.window is None

    # Assign the widget to a window
    widget.window = window
    assert len(window.widgets) == 1

    # Create a new window
    new_window = toga.Window()

    # Assign the widget to the new window
    widget.window = new_window

    # Window has been assigned
    assert widget.window == new_window

    # Window Widget registry has been updated
    assert len(window.widgets) == 0
    assert len(new_window.widgets) == 1
    assert new_window.widgets["widget_id"] == widget


def test_unset_window(widget):
    """A widget can be assigned to no window."""
    window = toga.Window()
    assert len(window.widgets) == 0
    assert widget.window is None

    # Assign the widget to a window
    widget.window = window
    assert len(window.widgets) == 1

    # Assign the widget to no window
    widget.window = None

    # The widget doesn't have a window
    assert widget.window is None

    # Window Widget registry has been updated
    assert len(window.widgets) == 0


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, False),
        ("", False),
        ("true", True),
        ("false", True),  # Evaluated as a string, this value is true.
        (0, False),
        (1234, True),
    ],
)
def test_enabled(widget, value, expected):
    """The enabled status of the widget can be changed."""
    # Widget is initially enabled by default.
    assert widget.enabled

    # Set the enabled status
    widget.enabled = value
    assert widget.enabled == expected

    # Disable the widget
    widget.enabled = False
    assert not widget.enabled

    # Set the enabled status again
    widget.enabled = value
    assert widget.enabled == expected


def test_refresh_root(widget):
    """Refresh can be invoked on the root node."""
    # Add children to the widget
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")
    widget.add(child1, child2, child3)

    # Refresh the root node
    widget.refresh()

    # Root widget was refreshed
    assert_action_performed(widget, "refresh")


def test_refresh_child(widget):
    """Refresh can be invoked on child."""
    # Add children to the widget
    child1 = ExampleLeafWidget(id="child1_id")
    child2 = ExampleLeafWidget(id="child2_id")
    child3 = ExampleLeafWidget(id="child3_id")
    widget.add(child1, child2, child3)

    # Refresh a child
    child2.refresh()

    # Child widget was refreshed
    assert_action_performed(child2, "refresh")

    # Root widget was refreshed
    assert_action_performed(widget, "refresh")


def test_focus(widget):
    """A widget can be given focus."""
    widget.focus()
    assert_action_performed(widget, "focus")


def test_tab_index(widget):
    """The tab index of a widget can be set and retrieved."""
    # The initial tab index is None
    assert widget.tab_index is None

    tab_index = 8
    widget.tab_index = tab_index

    assert widget.tab_index == 8
    assert attribute_value(widget, "tab_index") == tab_index



import pytest

import toga
from toga.colors import rgb
from toga.constants import Baseline, FillRule
from toga.fonts import SYSTEM, SYSTEM_DEFAULT_FONT_SIZE, Font
from toga.widgets.canvas import ClosedPathContext, Context, FillContext, StrokeContext
from toga_dummy.utils import assert_action_not_performed, assert_action_performed

REBECCA_PURPLE_COLOR = rgb(102, 51, 153)


def test_widget_created():
    """An empty canvas can be created."""
    widget = toga.Canvas()
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create Canvas")

    assert widget.on_resize._raw is None
    assert widget.on_press._raw is None
    assert widget.on_activate._raw is None
    assert widget.on_release._raw is None
    assert widget.on_drag._raw is None
    assert widget.on_alt_press._raw is None
    assert widget.on_alt_release._raw is None
    assert widget.on_alt_drag._raw is None

    # Canvas has a root context
    assert isinstance(widget.context, Context)


def test_create_with_value(
    widget,
    on_resize_handler,
    on_press_handler,
    on_activate_handler,
    on_release_handler,
    on_drag_handler,
    on_alt_press_handler,
    on_alt_release_handler,
    on_alt_drag_handler,
):
    """A Canvas can be created with initial values."""
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create Canvas")

    assert widget.on_resize._raw == on_resize_handler
    assert widget.on_press._raw == on_press_handler
    assert widget.on_activate._raw == on_activate_handler
    assert widget.on_release._raw == on_release_handler
    assert widget.on_drag._raw == on_drag_handler
    assert widget.on_alt_press._raw == on_alt_press_handler
    assert widget.on_alt_release._raw == on_alt_release_handler
    assert widget.on_alt_drag._raw == on_alt_drag_handler

    # Canvas has a root context
    assert isinstance(widget.context, Context)


def test_disable_no_op(widget):
    """Canvas doesn't have a disabled state."""
    # Enabled by default
    assert widget.enabled

    # Try to disable the widget
    widget.enabled = False

    # Still enabled.
    assert widget.enabled


def test_focus_noop(widget):
    """Focus is a no-op."""

    widget.focus()
    assert_action_not_performed(widget, "focus")


def test_redraw(widget):
    """The canvas can be redrawn."""
    widget.redraw()

    assert_action_performed(widget, "redraw")

    # An empty canvas has 2 draw operations - pushing and popping the root context.
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("pop context", {}),
    ]


def test_subcontext(widget):
    """A canvas can produce a subcontext."""
    with widget.Context() as subcontext:
        # A fresh context has been created as a subcontext of the canvas.
        assert isinstance(subcontext, Context)
        assert subcontext != widget.context


def test_closed_path(widget):
    """A canvas can produce a ClosedPath subcontext."""
    with widget.ClosedPath(x=10, y=20) as closed_path:
        # A fresh context has been created as a subcontext of the canvas.
        assert isinstance(closed_path, ClosedPathContext)
        assert closed_path != widget.context
        assert closed_path.x == 10
        assert closed_path.y == 20


def test_fill(widget):
    """A canvas can produce a Fill subcontext."""
    with widget.Fill(
        x=10, y=20, color="rebeccapurple", fill_rule=FillRule.EVENODD
    ) as fill:
        # A fresh context has been created as a subcontext of the canvas.
        assert isinstance(fill, FillContext)
        assert fill != widget.context

        assert fill.x == 10
        assert fill.y == 20
        assert fill.color == REBECCA_PURPLE_COLOR
        assert fill.fill_rule == FillRule.EVENODD


def test_stroke(widget):
    """A canvas can produce a Stroke subcontext."""
    with widget.Stroke(
        x=10, y=20, color="rebeccapurple", line_width=5, line_dash=[2, 7]
    ) as stroke:
        # A fresh context has been created as a subcontext of the canvas.
        assert isinstance(stroke, StrokeContext)
        assert stroke != widget.context

        assert stroke.x == 10
        assert stroke.y == 20
        assert stroke.color == REBECCA_PURPLE_COLOR
        assert stroke.line_width == 5.0
        assert stroke.line_dash == [2, 7]


@pytest.mark.parametrize(
    "font, expected",
    [
        (None, (132, 12)),
        (Font(family=SYSTEM, size=SYSTEM_DEFAULT_FONT_SIZE), (132, 12)),
        (Font(family=SYSTEM, size=20), (220, 20)),
        (Font(family="Cutive", size=SYSTEM_DEFAULT_FONT_SIZE), (198, 18)),
        (Font(family="Cutive", size=20), (330, 30)),
    ],
)
def test_measure_text(widget, font, expected):
    """Canvas can measure rendered text size."""
    assert widget.measure_text("Hello world", font=font) == expected


def test_as_image(widget):
    """A rendered canvas can be retrieved as an image."""
    image = widget.as_image()
    assert image is not None
    assert_action_performed(widget, "get image data")


def test_deprecated_drawing_operations(widget):
    """Deprecated simple drawing operations raise a warning."""

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.begin_path()",
    ):
        widget.new_path()

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.move_to()",
    ):
        widget.move_to(10, 20)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.line_to()",
    ):
        widget.line_to(10, 20)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.bezier_curve_to()",
    ):
        widget.bezier_curve_to(1, 2, 3, 4, 10, 20)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.quadratic_curve_to()",
    ):
        widget.quadratic_curve_to(1, 2, 10, 20)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.arc()",
    ):
        widget.arc(10, 20, 30, 0.4, 0.5, True)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.ellipse()",
    ):
        widget.ellipse(10, 20, 30, 40, 0.4, 0.5, 0.6, True)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.rect()",
    ):
        widget.rect(10, 20, 30, 40)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.write_text()",
    ):
        widget.write_text("Hello World", 10, 20, Font("Cutive", 37))

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.rotate()",
    ):
        widget.rotate(0.4)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.scale()",
    ):
        widget.scale(0.4, 0.5)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.translate()",
    ):
        widget.translate(10, 20)

    with pytest.warns(
        DeprecationWarning,
        match=r"Direct canvas operations have been deprecated; use context.reset_transform()",
    ):
        widget.reset_transform()

    with pytest.warns(
        DeprecationWarning,
        match=r"Canvas.closed_path\(\) has been renamed Canvas.ClosedPath\(\)",
    ):
        with widget.closed_path(10, 20) as closed_path:
            closed_path.line_to(20, 30)
            closed_path.line_to(30, 20)

    with pytest.warns(
        DeprecationWarning,
        match=r"Canvas.fill\(\) has been renamed Canvas.Fill\(\)",
    ):
        with widget.fill("rebeccapurple", FillRule.EVENODD) as fill:
            fill.line_to(20, 30)
            fill.line_to(30, 20)

    with pytest.warns(
        DeprecationWarning,
        match=r"Canvas.stroke\(\) has been renamed Canvas.Stroke\(\)",
    ):
        with widget.stroke("rebeccapurple", 3.0, [2, 7]) as stroke:
            stroke.line_to(20, 30)
            stroke.line_to(30, 20)

    # Assert the deprecated draw instructions rendered as expected
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("begin path", {}),
        ("move to", {"x": 10, "y": 20}),
        ("line to", {"x": 10, "y": 20}),
        (
            "bezier curve to",
            {"cp1x": 1, "cp1y": 2, "cp2x": 3, "cp2y": 4, "x": 10, "y": 20},
        ),
        (
            "quadratic curve to",
            {"cpx": 1, "cpy": 2, "x": 10, "y": 20},
        ),
        (
            "arc",
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "startangle": 0.4,
                "endangle": 0.5,
                "anticlockwise": True,
            },
        ),
        (
            "ellipse",
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": 0.4,
                "startangle": 0.5,
                "endangle": 0.6,
                "anticlockwise": True,
            },
        ),
        ("rect", {"x": 10, "y": 20, "width": 30, "height": 40}),
        (
            "write text",
            {
                "text": "Hello World",
                "x": 10,
                "y": 20,
                "font": Font("Cutive", 37)._impl,
                "baseline": Baseline.ALPHABETIC,
            },
        ),
        ("rotate", {"radians": 0.4}),
        ("scale", {"sx": 0.4, "sy": 0.5}),
        ("translate", {"tx": 10, "ty": 20}),
        ("reset transform", {}),
        # ClosedPath
        ("push context", {}),
        ("begin path", {}),
        ("move to", {"x": 10, "y": 20}),
        ("line to", {"x": 20, "y": 30}),
        ("line to", {"x": 30, "y": 20}),
        ("close path", {}),
        ("pop context", {}),
        # Fill
        ("push context", {}),
        ("begin path", {}),
        (
            "line to",
            {
                "x": 20,
                "y": 30,
                "fill_color": REBECCA_PURPLE_COLOR,
                "fill_rule": FillRule.EVENODD,
            },
        ),
        (
            "line to",
            {
                "x": 30,
                "y": 20,
                "fill_color": REBECCA_PURPLE_COLOR,
                "fill_rule": FillRule.EVENODD,
            },
        ),
        ("fill", {"color": REBECCA_PURPLE_COLOR, "fill_rule": FillRule.EVENODD}),
        ("pop context", {}),
        # Stroke
        ("push context", {}),
        ("begin path", {}),
        (
            "line to",
            {
                "x": 20,
                "y": 30,
                "stroke_color": REBECCA_PURPLE_COLOR,
                "line_width": 3.0,
                "line_dash": [2, 7],
            },
        ),
        (
            "line to",
            {
                "x": 30,
                "y": 20,
                "stroke_color": REBECCA_PURPLE_COLOR,
                "line_width": 3.0,
                "line_dash": [2, 7],
            },
        ),
        (
            "stroke",
            {"color": REBECCA_PURPLE_COLOR, "line_width": 3.0, "line_dash": [2, 7]},
        ),
        ("pop context", {}),
        ("pop context", {}),
    ]


def test_deprecated_args(widget):
    """Deprecated arguments to canvas functions raise warnings."""
    with pytest.warns(
        DeprecationWarning,
        match=r"The `tight` argument on Canvas.measure_text\(\) has been deprecated.",
    ):
        assert widget.measure_text(
            "Hello world", font=Font(family="Cutive", size=42), tight=True
        ) == (693, 63)

    with pytest.warns(
        DeprecationWarning,
        match=r"Canvas.fill\(\) has been renamed Canvas.Fill\(\)",
    ), pytest.warns(
        DeprecationWarning,
        match=r"The `preserve` argument on fill\(\) has been deprecated.",
    ):
        with widget.fill("rebeccapurple", FillRule.EVENODD, preserve=False) as fill:
            fill.line_to(20, 30)
            fill.line_to(30, 20)

    widget._impl.draw_instructions == [
        ("push context", {}),
        ("push context", {}),
        ("begin path", {}),
        ("line to", {"x": 20, "y": 30}),
        ("line to", {"x": 30, "y": 20}),
        ("fill", {"color": REBECCA_PURPLE_COLOR, "fill_rule": FillRule.EVENODD}),
        ("pop context", {}),
        ("pop context", {}),
    ]



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import Any, Protocol

import toga
from toga.handlers import wrapped_handler

from .base import StyleT, Widget


class OnChangeHandler(Protocol):
    def __call__(self, widget: TextInput, /, **kwargs: Any) -> object:
        """A handler to invoke when the text input is changed.

        :param widget: The TextInput that was changed.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class OnConfirmHandler(Protocol):
    def __call__(self, widget: TextInput, /, **kwargs: Any) -> object:
        """A handler to invoke when the text input is confirmed.

        :param widget: The TextInput that was confirmed.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class OnGainFocusHandler(Protocol):
    def __call__(self, widget: TextInput, /, **kwargs: Any) -> object:
        """A handler to invoke when the text input gains focus.

        :param widget: The TextInput that gained focus.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class OnLoseFocusHandler(Protocol):
    def __call__(self, widget: TextInput, /, **kwargs: Any) -> object:
        """A handler to invoke when the text input loses focus.

        :param widget: The TextInput that lost focus.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class TextInput(Widget):
    """Create a new single-line text input widget."""

    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        value: str | None = None,
        readonly: bool = False,
        placeholder: str | None = None,
        on_change: toga.widgets.textinput.OnChangeHandler | None = None,
        on_confirm: OnConfirmHandler | None = None,
        on_gain_focus: OnGainFocusHandler | None = None,
        on_lose_focus: OnLoseFocusHandler | None = None,
        validators: Iterable[Callable[[str], bool]] | None = None,
    ):
        """
        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param value: The initial content to display in the widget.
        :param readonly: Can the value of the widget be modified by the user?
        :param placeholder: The content to display as a placeholder when there is no
            user content to display.
        :param on_change: A handler that will be invoked when the value of the widget
            changes.
        :param on_confirm: A handler that will be invoked when the user accepts the
            value of the input (usually by pressing Return on the keyboard).
        :param on_gain_focus: A handler that will be invoked when the widget gains
            input focus.
        :param on_lose_focus: A handler that will be invoked when the widget loses
            input focus.
        :param validators: A list of validators to run on the value of the input.
        """
        super().__init__(id=id, style=style)

        # Create a platform specific implementation of the widget
        self._create()

        self.placeholder = placeholder
        self.readonly = readonly

        # Set the actual value before on_change, because we do not want
        # on_change triggered by it However, we need to prime the handler
        # property in case it is accessed.
        self.on_change = None
        self.on_confirm = None

        # Set the list of validators before we set the initial value so that
        # validation is performed on the initial value
        self.validators = validators
        self.value = value

        self.on_change = on_change
        self.on_confirm = on_confirm
        self.on_lose_focus = on_lose_focus
        self.on_gain_focus = on_gain_focus

    def _create(self) -> None:
        self._impl = self.factory.TextInput(interface=self)

    @property
    def readonly(self) -> bool:
        """Can the value of the widget be modified by the user?

        This only controls manual changes by the user (i.e., typing at the
        keyboard). Programmatic changes are permitted while the widget has
        ``readonly`` enabled.
        """
        return self._impl.get_readonly()

    @readonly.setter
    def readonly(self, value: object) -> None:
        self._impl.set_readonly(bool(value))

    @property
    def placeholder(self) -> str:
        """The placeholder text for the widget.

        A value of ``None`` will be interpreted and returned as an empty string.
        Any other object will be converted to a string using ``str()``.
        """
        return self._impl.get_placeholder()

    @placeholder.setter
    def placeholder(self, value: object) -> None:
        self._impl.set_placeholder("" if value is None else str(value))
        self.refresh()

    @property
    def value(self) -> str:
        """The text to display in the widget.

        A value of ``None`` will be interpreted and returned as an empty string.
        Any other object will be converted to a string using ``str()``.

        Any newline (``\\n``) characters in the string will be replaced with a space.

        Validation will be performed as a result of changing widget value.
        """
        return self._impl.get_value()

    @value.setter
    def value(self, value: object) -> None:
        if value is None:
            v = ""
        else:
            v = str(value).replace("\n", " ")
        self._impl.set_value(v)
        self.refresh()

    @property
    def is_valid(self) -> bool:
        """Does the value of the widget currently pass all validators without error?"""
        return self._impl.is_valid()

    @property
    def on_change(self) -> OnChangeHandler:
        """The handler to invoke when the value of the widget changes."""
        return self._on_change

    @on_change.setter
    def on_change(self, handler: toga.widgets.textinput.OnChangeHandler) -> None:
        self._on_change = wrapped_handler(self, handler)

    @property
    def validators(self) -> list[Callable[[str], bool]]:
        """The list of validators being used to check input on the widget.

        Changing the list of validators will cause validation to be performed.
        """
        return self._validators

    @validators.setter
    def validators(self, validators: Iterable[Callable[[str], bool]] | None) -> None:
        replacing = hasattr(self, "_validators")
        if validators is None:
            self._validators = []
        else:
            self._validators = list(validators)

        if replacing:
            self._validate()

    @property
    def on_gain_focus(self) -> OnGainFocusHandler:
        """The handler to invoke when the widget gains input focus."""
        return self._on_gain_focus

    @on_gain_focus.setter
    def on_gain_focus(self, handler: OnGainFocusHandler) -> None:
        self._on_gain_focus = wrapped_handler(self, handler)

    @property
    def on_lose_focus(self) -> OnLoseFocusHandler:
        """The handler to invoke when the widget loses input focus."""
        return self._on_lose_focus

    @on_lose_focus.setter
    def on_lose_focus(self, handler: OnLoseFocusHandler) -> None:
        self._on_lose_focus = wrapped_handler(self, handler)

    def _validate(self) -> None:
        """Validate the current value of the widget.

        If a problem is found, the widget will be put into an error state.
        """
        for validator in self.validators:
            error_message = validator(self.value)
            if error_message is not None:
                self._impl.set_error(error_message)
                break
        else:
            self._impl.clear_error()

    def _value_changed(self) -> None:
        """Validate the current value of the widget and invoke the on_change handler."""
        self._validate()
        self.on_change()

    @property
    def on_confirm(self) -> OnConfirmHandler:
        """The handler to invoke when the user accepts the value of the widget,
        usually by pressing return/enter on the keyboard.
        """
        return self._on_confirm

    @on_confirm.setter
    def on_confirm(self, handler: OnConfirmHandler) -> None:
        self._on_confirm = wrapped_handler(self, handler)



from __future__ import annotations

from typing import Literal, SupportsFloat

from .base import StyleT, Widget


class ProgressBar(Widget):
    _MIN_WIDTH = 100

    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        max: str | SupportsFloat = 1.0,
        value: str | SupportsFloat = 0.0,
        running: bool = False,
    ):
        """Create a new Progress Bar widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param max: The value that represents completion of the task. Must
            be > 0.0; defaults to 1.0. A value of ``None`` indicates that the task
            length is indeterminate.
        :param value: The current progress against the maximum value. Must be
            between 0.0 and ``max``; any value outside this range will be
            clipped. Defaults to 0.0.
        :param running: Describes whether the indicator is running at the time
            it is created. Default is False.
        """
        super().__init__(id=id, style=style)

        self._impl = self.factory.ProgressBar(interface=self)

        self.max = max
        self.value = value

        if running:
            self.start()

    @property
    def enabled(self) -> Literal[True]:
        """Is the widget currently enabled? i.e., can the user interact with the widget?

        ProgressBar widgets cannot be disabled; this property will always return True;
        any attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value: object) -> None:
        pass

    @property
    def is_running(self) -> bool:
        """Describe if the activity indicator is currently running.

        Use ``start()`` and ``stop()`` to change the running state.

        True if this activity indicator is running; False otherwise.
        """
        return self._impl.is_running()

    @property
    def is_determinate(self) -> bool:
        """Describe whether the progress bar has a known or indeterminate maximum.

        True if the progress bar has determinate length; False otherwise.
        """
        return self.max is not None

    def start(self) -> None:
        """Start the progress bar.

        If the progress bar is already started, this is a no-op.
        """
        if not self.is_running:
            self._impl.start()

    def stop(self) -> None:
        """Stop the progress bar.

        If the progress bar is already stopped, this is a no-op.
        """
        if self.is_running:
            self._impl.stop()

    @property
    def value(self) -> float:
        """The current value of the progress indicator.

        If the progress bar is determinate, the value must be between 0 and
        ``max``. Any value outside this range will be clipped.

        If the progress bar is indeterminate, changes in value will be ignored,
        and the current value will be returned as ``None``.
        """
        return self._impl.get_value()

    @value.setter
    def value(self, value: str | SupportsFloat) -> None:
        if self.max is not None:
            value = max(0.0, min(self.max, float(value)))
            self._impl.set_value(value)

    @property
    def max(self) -> float | None:
        """The value indicating completion of the task being monitored.

        Must be a number > 0, or ``None`` for a task of indeterminate length.
        """
        return self._impl.get_max()

    @max.setter
    def max(self, value: str | SupportsFloat | None) -> None:
        if value is None:
            self._impl.set_max(None)
        elif float(value) > 0.0:
            self._impl.set_max(float(value))
        else:
            raise ValueError("max value must be None, or a numerical value > 0")



from unittest.mock import Mock

import toga
from toga_dummy.utils import assert_action_performed


def test_widget_created():
    """A text input can be created."""
    widget = toga.PasswordInput()
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create PasswordInput")

    assert not widget.readonly
    assert widget.placeholder == ""
    assert widget.value == ""
    assert widget._on_change._raw is None
    assert widget._on_confirm._raw is None
    assert widget._on_gain_focus._raw is None
    assert widget._on_lose_focus._raw is None
    assert widget.validators == []


def test_create_with_values():
    """A multiline text input can be created with initial values."""
    on_change = Mock()
    on_confirm = Mock()
    on_gain_focus = Mock()
    on_lose_focus = Mock()
    validator1 = Mock(return_value=None)
    validator2 = Mock(return_value=None)

    widget = toga.PasswordInput(
        value="Some text",
        placeholder="A placeholder",
        readonly=True,
        on_change=on_change,
        on_confirm=on_confirm,
        on_gain_focus=on_gain_focus,
        on_lose_focus=on_lose_focus,
        validators=[validator1, validator2],
    )
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create PasswordInput")

    assert widget.readonly
    assert widget.placeholder == "A placeholder"
    assert widget.value == "Some text"
    assert widget._on_change._raw == on_change
    assert widget._on_confirm._raw == on_confirm
    assert widget._on_gain_focus._raw == on_gain_focus
    assert widget._on_lose_focus._raw == on_lose_focus
    assert widget.validators == [validator1, validator2]

    # Validators have been invoked with the initial text
    validator1.assert_called_once_with("Some text")
    validator2.assert_called_once_with("Some text")

    # Change handler hasn't been invoked
    on_change.assert_not_called()



import pytest

import toga
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
)


@pytest.fixture
def activity_indicator():
    return toga.ActivityIndicator()


def test_widget_created(activity_indicator):
    """An activity indicator can be created."""
    # Round trip the impl/interface
    assert activity_indicator._impl.interface == activity_indicator
    assert_action_performed(activity_indicator, "create ActivityIndicator")


def test_disable_no_op(activity_indicator):
    """ActivityIndicator doesn't have a disabled state."""
    # Enabled by default
    assert activity_indicator.enabled

    # Try to disable the widget
    activity_indicator.enabled = False

    # Still enabled.
    assert activity_indicator.enabled


def test_start(activity_indicator):
    """An activity indicator can be started."""
    # Not running initially
    assert not activity_indicator.is_running

    # Assert that start was not invoked on the impl as part of creation.
    assert_action_not_performed(activity_indicator, "start ActivityIndicator")

    # Start spinning
    activity_indicator.start()

    # The indicator is now running
    assert activity_indicator.is_running

    # The impl was triggered.
    assert_action_performed(activity_indicator, "start ActivityIndicator")


def test_already_started(activity_indicator):
    """If an activity indicator is already started, starting again is a no-op."""
    # Start the activity indicator
    activity_indicator.start()

    # Reset the event log so we can detect new events
    EventLog.reset()

    # Start the indicator again
    activity_indicator.start()

    # The indicator is still running
    assert activity_indicator.is_running

    # No action was performed.
    assert_action_not_performed(activity_indicator, "start ActivityIndicator")


def test_stop(activity_indicator):
    """An indicator can be stopped."""
    # Start spinning
    activity_indicator.start()

    # The indicator is running
    assert activity_indicator.is_running

    # Stop spinning
    activity_indicator.stop()

    # The indicator is no longer running
    assert not activity_indicator.is_running

    # The impl has been stopped
    assert_action_performed(activity_indicator, "stop ActivityIndicator")


def test_already_stopped(activity_indicator):
    """If an indicator is already stopped, stopping again is a no-op."""
    # The indicator is not running initially
    assert not activity_indicator.is_running

    # Stop spinning
    activity_indicator.stop()

    # The indicator is still not running
    assert not activity_indicator.is_running

    # No stop action was performed.
    assert_action_not_performed(activity_indicator, "stop ActivityIndicator")


def test_initially_running():
    """An activity indicator can be created in a started state."""
    # Creating a new progress bar with running=True so it is already running
    activity_indicator = toga.ActivityIndicator(running=True)

    # Indicator is running
    assert activity_indicator.is_running

    # Assert that start was invoked on the impl as part of creation.
    assert_action_performed(activity_indicator, "start ActivityIndicator")


def test_focus_noop(activity_indicator):
    """Focus is a no-op."""

    activity_indicator.focus()
    assert_action_not_performed(activity_indicator, "focus")



from unittest.mock import Mock

import pytest

import toga
from toga_dummy.utils import EventLog, assert_action_performed, attribute_value


@pytest.fixture
def widget():
    return toga.MultilineTextInput()


def test_widget_created(widget):
    """A multiline text input."""
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create MultilineTextInput")

    assert not widget.readonly
    assert widget.placeholder == ""
    assert widget.value == ""
    assert widget._on_change._raw is None


def test_create_with_values():
    """A multiline text input can be created with initial values."""
    on_change = Mock()
    widget = toga.MultilineTextInput(
        value="Some text",
        placeholder="A placeholder",
        readonly=True,
        on_change=on_change,
    )
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create MultilineTextInput")

    assert widget.readonly
    assert widget.placeholder == "A placeholder"
    assert widget.value == "Some text"
    assert widget._on_change._raw == on_change

    # Change handler hasn't been invoked
    on_change.assert_not_called()


@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Text", "New Text"),
        ("", ""),
        (None, ""),
        (12345, "12345"),
        ("Contains\nnewline", "Contains\nnewline"),
    ],
)
def test_value(widget, value, expected):
    """The value of the input can be set."""
    # Clear the event log
    EventLog.reset()

    # Install an on_change handler
    handler = Mock()
    widget.on_change = handler

    widget.value = value
    assert widget.value == expected

    # test backend has the right value
    assert attribute_value(widget, "value") == expected

    # A refresh was performed
    assert_action_performed(widget, "refresh")

    # Callback was invoked
    handler.assert_called_once_with(widget)


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, False),
        ("", False),
        ("true", True),
        ("false", True),  # Evaluated as a string, this value is true.
        (0, False),
        (1234, True),
    ],
)
def test_readonly(widget, value, expected):
    """The readonly status of the widget can be changed."""
    # Widget is initially not readonly by default.
    assert not widget.readonly

    # Set the readonly status
    widget.readonly = value
    assert widget.readonly == expected

    # Set the widget readonly
    widget.readonly = True
    assert widget.readonly

    # Set the readonly status again
    widget.readonly = value
    assert widget.readonly == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Text", "New Text"),
        ("", ""),
        (None, ""),
        (12345, "12345"),
        ("Contains\nnewline", "Contains\nnewline"),
    ],
)
def test_placeholder(widget, value, expected):
    """The value of the placeholder can be set."""
    # Clear the event log
    EventLog.reset()

    widget.placeholder = value
    assert widget.placeholder == expected

    # test backend has the right value
    assert attribute_value(widget, "placeholder") == expected

    # A refresh was performed
    assert_action_performed(widget, "refresh")


def test_scroll(widget):
    """The widget can be scrolled programmatically."""
    # Clear the event log
    EventLog.reset()

    widget.scroll_to_top()

    # A refresh was performed
    assert_action_performed(widget, "scroll to top")

    # Clear the event log
    EventLog.reset()

    widget.scroll_to_bottom()

    # The widget has been scrolled
    assert_action_performed(widget, "scroll to bottom")


def test_on_change(widget):
    """The on_change handler can be invoked."""
    # No handler initially
    assert widget._on_change._raw is None

    # Define and set a new callback
    handler = Mock()

    widget.on_change = handler

    assert widget.on_change._raw == handler

    # Invoke the callback
    widget._impl.simulate_change()

    # Callback was invoked
    handler.assert_called_once_with(widget)



from pathlib import Path

import PIL.Image
import pytest

import toga
from toga_dummy.plugins.image_formats import (
    CustomImage,
    CustomImageSubclass,
    DisabledImageConverter,
)
from toga_dummy.utils import assert_action_performed_with

RELATIVE_FILE_PATH = Path("resources/sample.png")
ABSOLUTE_FILE_PATH = Path(__file__).parent / "resources/sample.png"


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.parametrize(
    "args, kwargs",
    [
        # Fully qualified path
        ((ABSOLUTE_FILE_PATH,), {}),
        ((), {"src": ABSOLUTE_FILE_PATH}),
        ((), {"path": ABSOLUTE_FILE_PATH}),
        # Fully qualified string
        ((f"{ABSOLUTE_FILE_PATH}",), {}),
        ((), {"src": f"{ABSOLUTE_FILE_PATH}"}),
        ((), {"path": f"{ABSOLUTE_FILE_PATH}"}),
        # Relative path
        ((RELATIVE_FILE_PATH,), {}),
        ((), {"src": RELATIVE_FILE_PATH}),
        ((), {"path": RELATIVE_FILE_PATH}),
        # Relative string
        ((f"{RELATIVE_FILE_PATH}",), {}),
        ((), {"src": f"{RELATIVE_FILE_PATH}"}),
        ((), {"path": f"{RELATIVE_FILE_PATH}"}),
    ],
)
def test_create_from_file(app, args, kwargs):
    """An image can be constructed from a file."""
    image = toga.Image(*args, **kwargs)

    # Image is bound
    assert image._impl is not None
    # impl/interface round trips
    assert image._impl.interface == image

    # The image's path is fully qualified
    assert image._impl.interface.path == ABSOLUTE_FILE_PATH


MISSING_ABSOLUTE_PATH = Path.home() / "does/not/exist/image.jpg"
MISSING_RELATIVE_PATH = Path("does/not/exist/image.jpg")


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.parametrize(
    "args, kwargs",
    [
        # Empty string
        (("",), {}),
        ((), {"src": ""}),
        # Absolute path
        ((MISSING_ABSOLUTE_PATH,), {}),
        ((), {"src": MISSING_ABSOLUTE_PATH}),
        ((), {"path": MISSING_ABSOLUTE_PATH}),
        # Absolute string
        ((f"{MISSING_ABSOLUTE_PATH}",), {}),
        ((), {"src": f"{MISSING_ABSOLUTE_PATH}"}),
        ((), {"path": f"{MISSING_ABSOLUTE_PATH}"}),
        # Relative path
        ((MISSING_RELATIVE_PATH,), {}),
        ((), {"src": f"{MISSING_RELATIVE_PATH}"}),
        ((), {"path": f"{MISSING_RELATIVE_PATH}"}),
        # Relative string
        ((f"{MISSING_RELATIVE_PATH}",), {}),
        ((), {"src": f"{MISSING_RELATIVE_PATH}"}),
        ((), {"path": f"{MISSING_RELATIVE_PATH}"}),
    ],
)
def test_create_with_nonexistent_file(app, args, kwargs):
    """If a file image source doesn't exist, an error is raised."""
    with pytest.raises(FileNotFoundError):
        toga.Image(*args, **kwargs)


BYTES = ABSOLUTE_FILE_PATH.read_bytes()


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.parametrize(
    "args, kwargs",
    [
        ((BYTES,), {}),
        ((), {"src": BYTES}),
        ((), {"data": BYTES}),
        # Other "lump of bytes" data types
        ((bytearray(BYTES),), {}),
        ((memoryview(BYTES),), {}),
    ],
)
def test_create_from_bytes(args, kwargs):
    """An image can be constructed from data."""
    image = toga.Image(*args, **kwargs)

    # Image is bound
    assert image._impl is not None
    # impl/interface round trips
    assert image._impl.interface == image

    # Image was constructed with data
    assert_action_performed_with(image, "load image data", data=BYTES)


def test_create_from_raw():
    """An image can be created from a raw data source."""
    orig = toga.Image(BYTES)

    copy = toga.Image(orig._impl.native)
    # Image is bound
    assert copy._impl is not None
    # impl/interface round trips
    assert copy._impl.interface == copy

    # Image was constructed from raw data
    assert_action_performed_with(copy, "load image from raw")


def test_no_source():
    """If no source is provided, an error is raised."""
    with pytest.raises(
        TypeError,
        match=r"Image.__init__\(\) missing 1 required positional argument: 'src'",
    ):
        toga.Image()


def test_empty_image():
    """If the image source is provided as None, an error is raised."""
    with pytest.raises(
        TypeError,
        match=r"Unsupported source type for Image",
    ):
        toga.Image(None)


def test_empty_image_explicit():
    """If src is explicitly provided as None, an error is raised."""
    with pytest.raises(
        TypeError,
        match=r"Unsupported source type for Image",
    ):
        toga.Image(src=None)


def test_invalid_input_format():
    """Trying to create an image with an invalid input should raise an error."""
    with pytest.raises(
        TypeError,
        match=r"Unsupported source type for Image",
    ):
        toga.Image(42)


def test_create_from_pil(app):
    """An image can be created from a PIL image."""
    with PIL.Image.open(ABSOLUTE_FILE_PATH) as pil_image:
        pil_image.load()
    toga_image = toga.Image(pil_image)

    assert isinstance(toga_image, toga.Image)
    assert toga_image.size == (144, 72)


def test_create_from_toga_image(app):
    """An image can be created from another Toga image."""
    toga_image = toga.Image(ABSOLUTE_FILE_PATH)
    toga_image_2 = toga.Image(toga_image)

    assert isinstance(toga_image_2, toga.Image)
    assert toga_image_2.size == (144, 72)


@pytest.mark.parametrize("kwargs", [{"data": BYTES}, {"path": ABSOLUTE_FILE_PATH}])
def test_deprecated_arguments(kwargs):
    with pytest.deprecated_call():
        toga.Image(**kwargs)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.parametrize(
    "args, kwargs",
    [
        # One positional, one keyword
        ((ABSOLUTE_FILE_PATH,), {"data": BYTES}),
        ((ABSOLUTE_FILE_PATH,), {"path": ABSOLUTE_FILE_PATH}),
        # Two keywords
        ((), {"data": BYTES, "path": ABSOLUTE_FILE_PATH}),
        # All three
        ((ABSOLUTE_FILE_PATH,), {"data": BYTES, "path": ABSOLUTE_FILE_PATH}),
    ],
)
def test_too_many_arguments(args, kwargs):
    """If multiple arguments are supplied, an error is raised."""
    with pytest.raises(
        TypeError,
        match=r"Received multiple arguments to constructor.",
    ):
        toga.Image(*args, **kwargs)


def test_dimensions(app):
    """The dimensions of the image can be retrieved."""
    image = toga.Image(RELATIVE_FILE_PATH)

    assert image.size == (144, 72)
    assert image.width == 144
    assert image.height == 72


def test_data(app):
    """The raw data of the image can be retrieved."""
    image = toga.Image(ABSOLUTE_FILE_PATH)

    # We can't guarantee the round-trip of image data,
    # but the data starts with a PNG header
    assert image.data.startswith(b"\x89PNG\r\n\x1a\n")

    # If we build a new image from the data, it has the same properties.
    from_data = toga.Image(image.data)
    assert from_data.width == image.width
    assert from_data.height == image.height


def test_image_save(tmp_path):
    """An image can be saved."""
    save_path = tmp_path / "save.png"
    image = toga.Image(BYTES)
    image.save(save_path)

    assert_action_performed_with(image, "save", path=save_path)


class ImageSubclass(toga.Image):
    pass


@pytest.mark.parametrize(
    "Class_1, Class_2",
    [
        (toga.Image, toga.Image),
        (toga.Image, ImageSubclass),
        (ImageSubclass, toga.Image),
        (ImageSubclass, ImageSubclass),
    ],
)
def test_as_format_toga(app, Class_1, Class_2):
    """as_format can successfully return a "copy" Image, with support for
    subclassing."""
    image_1 = Class_1(ABSOLUTE_FILE_PATH)
    image_2 = image_1.as_format(Class_2)

    assert isinstance(image_2, Class_2)
    assert image_2.size == (144, 72)


def test_as_format_pil(app):
    """as_format can successfully return a PIL image."""
    toga_image = toga.Image(ABSOLUTE_FILE_PATH)
    pil_image = toga_image.as_format(PIL.Image.Image)
    assert isinstance(pil_image, PIL.Image.Image)
    assert pil_image.size == (144, 72)


@pytest.mark.parametrize("ImageClass", [CustomImage, CustomImageSubclass])
def test_create_from_custom_class(app, ImageClass):
    """toga.Image can be created from custom type."""
    custom_image = ImageClass()
    toga_image = toga.Image(custom_image)
    assert isinstance(toga_image, toga.Image)
    assert toga_image.size == (144, 72)


@pytest.mark.parametrize("ImageClass", [CustomImage, CustomImageSubclass])
def test_as_format_custom_class(app, ImageClass):
    """as_format can successfully return a registered custom image type."""
    toga_image = toga.Image(ABSOLUTE_FILE_PATH)
    custom_image = toga_image.as_format(ImageClass)
    assert isinstance(custom_image, ImageClass)
    assert custom_image.size == (144, 72)


def test_disabled_image_plugin(app):
    """Disabled image plugin shouldn't be available."""
    assert DisabledImageConverter not in toga.Image._converters()


# None is same as supplying nothing; also test a random unrecognized class
@pytest.mark.parametrize("arg", [None, toga.Button])
def test_as_format_invalid_input(app, arg):
    """An unsupported format raises an error."""
    toga_image = toga.Image(ABSOLUTE_FILE_PATH)

    with pytest.raises(TypeError, match=r"Unknown conversion format for Image:"):
        toga_image.as_format(arg)



import pytest

import toga
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
)


def test_divider_created():
    """A divider can be created."""
    divider = toga.Divider()

    # Round trip the impl/interface
    assert divider._impl.interface == divider
    assert_action_performed(divider, "create Divider")

    # Default direction is honored
    assert divider.direction == toga.Divider.HORIZONTAL


@pytest.mark.parametrize("direction", [toga.Divider.HORIZONTAL, toga.Divider.VERTICAL])
def test_divider_created_explicit(direction):
    """A divider can be created."""
    divider = toga.Divider(direction=direction)

    # Round trip the impl/interface
    assert divider._impl.interface == divider
    assert_action_performed(divider, "create Divider")

    # Default direction is honored
    assert divider.direction == direction


def test_disable_no_op():
    """Divider doesn't have a disabled state."""
    divider = toga.Divider()

    # Enabled by default
    assert divider.enabled

    # Try to disable the widget
    divider.enabled = False

    # Still enabled.
    assert divider.enabled


def test_update_direction():
    """The direction of the divider can be altered."""
    divider = toga.Divider(direction=toga.Divider.HORIZONTAL)

    # Initial direction is as expected
    assert divider.direction == toga.Divider.HORIZONTAL

    # Reset the event log.
    EventLog.reset()

    # Change the direction.
    divider.direction = toga.Divider.VERTICAL

    # The direction has been changed, and a refresh requested
    assert divider.direction == toga.Divider.VERTICAL
    assert_action_performed(divider, "refresh")


def test_focus_noop():
    """Focus is a no-op."""
    divider = toga.Divider(direction=toga.Divider.HORIZONTAL)

    divider.focus()
    assert_action_not_performed(divider, "focus")



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, NamedTuple

import toga

if TYPE_CHECKING:
    if sys.version_info < (3, 10):
        from typing_extensions import TypeAlias
    else:
        from typing import TypeAlias

    PositionT: TypeAlias = toga.Position | tuple[int, int]
    SizeT: TypeAlias = toga.Size | tuple[int, int]


class LatLng(NamedTuple):
    """A geographic coordinate."""

    #: Latitude
    lat: float

    #: Longitude
    lng: float

    def __str__(self) -> str:
        return f"({self.lat:6f}, {self.lng:6f})"


class Position(NamedTuple):
    """A 2D window position."""

    #: X coordinate, in CSS pixels.
    x: int

    #: Y coordinate, in CSS pixels.
    y: int

    def __str__(self) -> str:
        return f"({self.x}, {self.y})"

    def __add__(self, other):
        return Position(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Position(self.x - other.x, self.y - other.y)


class Size(NamedTuple):
    """A 2D window size."""

    #: Width
    width: int

    #: Height
    height: int

    def __str__(self) -> str:
        return f"({self.width} x {self.height})"



from __future__ import annotations

from builtins import id as identifier
from typing import TYPE_CHECKING, Any, TypeVar

from travertino.declaration import BaseStyle
from travertino.node import Node

from toga.platform import get_platform_factory
from toga.style import Pack, TogaApplicator

if TYPE_CHECKING:
    from toga.app import App
    from toga.window import Window

StyleT = TypeVar("StyleT", bound=BaseStyle)


class Widget(Node):
    _MIN_WIDTH = 100
    _MIN_HEIGHT = 100

    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
    ):
        """Create a base Toga widget.

        This is an abstract base class; it cannot be instantiated.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        """
        super().__init__(
            style=style if style else Pack(),
            applicator=TogaApplicator(self),
        )

        self._id = str(id if id else identifier(self))
        self._window: Window | None = None
        self._app: App | None = None
        self._impl: Any = None

        self.factory = get_platform_factory()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}:0x{identifier(self):x}>"

    def __lt__(self, other: Widget) -> bool:
        return self.id < other.id

    @property
    def id(self) -> str:
        """The DOM identifier for the widget.

        This id can be used to target CSS directives.
        """
        return self._id

    @property
    def tab_index(self) -> int | None:
        """The position of the widget in the focus chain for the window.

        .. note::

            This is a beta feature. The ``tab_index`` API may change in the future.
        """
        return self._impl.get_tab_index()

    @tab_index.setter
    def tab_index(self, tab_index: int) -> None:
        self._impl.set_tab_index(tab_index)

    def _assert_can_have_children(self) -> None:
        if not self.can_have_children:
            raise ValueError(f"{type(self).__name__} cannot have children")

    def add(self, *children: Widget) -> None:
        """Add the provided widgets as children of this widget.

        If a child widget already has a parent, it will be re-parented as a
        child of this widget. If the child widget is already a child of this
        widget, there is no change.

        :param children: The widgets to add as children of this widget.
        :raises ValueError: If this widget cannot have children.
        """
        self._assert_can_have_children()
        for child in children:
            if child.parent is not self:
                # remove from old parent
                if child.parent:
                    child.parent.remove(child)

                # Set app and window. This is done *before* changing any parenting
                # relationships, so that the widget registry can verify the widget ID is
                # unique. App must be set before window to ensure the widget registry
                # can be found.
                child.app = self.app
                child.window = self.window

                # add to new parent
                super().add(child)

                self._impl.add_child(child._impl)

        # Whatever layout we're a part of needs to be refreshed
        self.refresh()

    def insert(self, index: int, child: Widget) -> None:
        """Insert a widget as a child of this widget.

        If a child widget already has a parent, it will be re-parented as a
        child of this widget. If the child widget is already a child of this
        widget, there is no change.

        :param index: The position in the list of children where the new widget
            should be added.
        :param child: The child to insert as a child of this node.
        :raises ValueError: If this widget cannot have children.
        """
        self._assert_can_have_children()
        if child.parent is not self:
            # remove from old parent
            if child.parent:
                child.parent.remove(child)

            # Set app and window. This is done *before* changing any parenting
            # relationships, so that the widget registry can verify the widget ID is
            # unique. App must be set before window to ensure the widget registry
            # can be found.
            child.app = self.app
            child.window = self.window

            # add to new parent
            super().insert(index, child)

            self._impl.insert_child(index, child._impl)

        # Whatever layout we're a part of needs to be refreshed
        self.refresh()

    def index(self, child: Widget) -> int:
        """Get the index of a widget in the list of children of this widget.

        :param child: The child widget of interest.
        :raises ValueError: If the specified child widget is not found in the
            list of children.

        :returns: Index of specified child widget in children list.
        """
        for _ind, _child in enumerate(self._children):
            if child == _child:
                return _ind
        raise ValueError(f"{type(child).__name__} not found")

    def replace(self, old_child: Widget, new_child: Widget) -> None:
        """Replace an existing child widget with a new child widget.

        :param old_child: The existing child widget to be replaced.
        :param new_child: The new child widget to be included.
        """
        old_child_index = self.index(old_child)
        self.remove(old_child)
        self.insert(old_child_index, new_child)

    def remove(self, *children: Widget) -> None:
        """Remove the provided widgets as children of this node.

        Any nominated child widget that is not a child of this widget will
        not have any change in parentage.

        Refreshes the widget after removal if any children were removed.

        :param children: The child nodes to remove.
        :raises ValueError: If this widget cannot have children.
        """
        self._assert_can_have_children()

        removed = False
        for child in children:
            if child.parent is self:
                removed = True
                super().remove(child)

                # Remove from the window before removing from the app
                # so that the widget can be removed from the app-level registry.
                child.window = None
                child.app = None

                self._impl.remove_child(child._impl)

        # If we removed something, whatever layout we're a part of needs to be refreshed
        if removed:
            self.refresh()

    def clear(self) -> None:
        """Remove all child widgets of this node.

        Refreshes the widget after removal if any children were removed.

        :raises ValueError: If this widget cannot have children.
        """
        self._assert_can_have_children()
        self.remove(*self.children)

    @property
    def app(self) -> App | None:
        """The App to which this widget belongs.

        When setting the app for a widget, all children of this widget will be
        recursively assigned to the same app.

        :raises ValueError: If this widget is already associated with another app.
        """
        return self._app

    @app.setter
    def app(self, app: App | None) -> None:
        # If the widget is already assigned to an app
        if self._app:
            if self._app == app:
                # If app is the same as the previous app, return
                return

        self._app = app
        self._impl.set_app(app)
        for child in self.children:
            child.app = app

    @property
    def window(self) -> Window | None:
        """The window to which this widget belongs.

        When setting the window for a widget, all children of this widget will be
        recursively assigned to the same window.

        If the widget has a value for :any:`window`, it *must* also have a value for
        :any:`app`.
        """
        return self._window

    @window.setter
    def window(self, window: Window | None) -> None:
        if self.window is not None and window is None:
            # If the widget is currently in the registry, but is being removed from a
            # window, remove the widget from the widget registry
            self.window.app.widgets._remove(self.id)
        elif self.window is None and window is not None:
            # If the widget is being assigned to a window for the first time, add it to the widget registry
            window.app.widgets._add(self)

        self._window = window
        self._impl.set_window(window)

        for child in self.children:
            child.window = window

    @property
    def enabled(self) -> bool:
        """Is the widget currently enabled? i.e., can the user interact with the widget?"""
        return self._impl.get_enabled()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._impl.set_enabled(bool(value))

    def refresh(self) -> None:
        self._impl.refresh()

        # Refresh the layout
        if self._root:
            # We're not the root of the node hierarchy;
            # defer the refresh call to the root node.
            self._root.refresh()
        else:
            # We can't compute a layout until we have a container
            if self._impl.container:
                super().refresh(self._impl.container)
                self._impl.container.refreshed()

    def focus(self) -> None:
        """Give this widget the input focus.

        This method is a no-op if the widget can't accept focus. The ability of a widget
        to accept focus is platform-dependent. In general, on desktop platforms you can
        focus any widget that can accept user input, while on mobile platforms focus is
        limited to widgets that accept text input (i.e., widgets that cause the virtual
        keyboard to appear).
        """
        self._impl.focus()



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from toga.images import BytesLikeT

# Presumably, other converter plugins will be included with, or only installed
# alongside, the packages they're for. But since this is provided in Toga, we need to
# check if Pillow is actually installed, and disable this plugin otherwise.
try:
    import PIL.Image

    PIL_imported = True

except ImportError:  # pragma: no cover
    PIL_imported = False


class PILConverter:
    image_class = PIL.Image.Image if PIL_imported else None

    @staticmethod
    def convert_from_format(image_in_format: PIL.Image.Image) -> bytes:
        buffer = BytesIO()
        image_in_format.save(buffer, format="png", compress_level=0)
        return buffer.getvalue()

    @staticmethod
    def convert_to_format(
        data: BytesLikeT,
        image_class: type[PIL.Image.Image],
    ) -> PIL.Image.Image:
        # PIL Images aren't designed to be subclassed, so no implementation is necessary
        # for a supplied format.
        buffer = BytesIO(data)
        with PIL.Image.open(buffer) as pil_image:
            pil_image.load()
        return pil_image
